#!/usr/bin/env python

__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "MIT"


'''This examples shows how to use the rs.Filesystem API
   with the SFTP file adaptor.

   If something doesn't work as expected, try to set
   SAGA_VERBOSE=3 in your environment before you run the
   script in order to get some debug output.

   If you think you have encountered a defect, please 
   report it at: https://github.com/radical-cybertools/radical.saga/issues
'''

import sys
import getpass

import radical.saga as rs


def main():

    try:
        # Your ssh identity on the remote machine.
        ctx = rs.Context("ssh")
       #ctx.user_id = getpass.getuser()  # Change if necessary

        session = rs.Session()
        session.add_context(ctx)

        # open home directory on a remote machine
        remote_dir = rs.filesystem.Directory('sftp://stampede.tacc.xsede.org/tmp/',
                                               session=session)

        for entry in remote_dir.list():
            if remote_dir.is_dir(entry):
                print("d %12s %s" % (remote_dir.get_size(entry), entry))
            else:
                print("- %12s %s" % (remote_dir.get_size(entry), entry))
        return 0

    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (ex.type, (str(ex))))
        # Trace back the exception. That can be helpful for debugging.
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        return -1

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

__author__    = "Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" This examples shows how to run a job on a remote PBS/TORQUE cluster
    using the 'PBS' job adaptor.

    More information about the radical.saga job API can be found at:
    http://radical-cybertools.github.com/radical.saga/doc/library/job/index.html
"""

import sys

import radical.saga as rs


# ----------------------------------------------------------------------------
# This is an example for a callback function. Callback functions can be
# registered with a rs.Job object and get 'fired' asynchronously on
# certain conditions.
def job_state_change_cb(src_obj, fire_on, value):
    print("Callback    : job state changed to '%s'\n" % value)
    return True


# ----------------------------------------------------------------------------
#
def main():

    try:
        # Your ssh identity on the remote machine.
        ctx = rs.Context("ssh")

        # Change e.g., if you have a differnent username on the remote machine
        #ctx.user_id = "your_ssh_username"

        session = rs.Session()
        session.add_context(ctx)

        # Create a job service object that represent a remote pbs cluster.
        # The keyword 'pbs' in the url scheme triggers the PBS adaptors
        # and '+ssh' enables PBS remote access via SSH.
        js = rs.job.Service("ssh://localhost",
                              session=session)

        # Next, we describe the job we want to run. A complete set of job
        # description attributes can be found in the API documentation.
        jd = rs.job.Description()
        jd.environment       = {'FILENAME': 'testfile'}
        jd.wall_time_limit   = 1 # minutes

        jd.executable        = '/bin/touch'
        jd.arguments         = ['$FILENAME']

        #jd.total_cpu_count   = 12 # for lonestar this has to be a multiple of 12
        #jd.spmd_variation    = '12way' # translates to the qsub -pe flag

      # jd.queue             = "batch"
      # jd.project           = "TG-MCB090174"

        jd.working_directory = "$HOME/A/B/C"
        jd.output            = "examplejob.out"
        jd.error             = "examplejob.err"

        # Create a new job from the job description. The initial state of
        # the job is 'New'.
        touchjob = js.create_job(jd)

        # Register our callback. We want it to 'fire' on job state change
        touchjob.add_callback(rs.STATE, job_state_change_cb)

        # Check our job's id and state
        print("Job ID      : %s" % (touchjob.id))
        print("Job State   : %s" % (touchjob.state))

        # Now we can start our job.
        print("\n...starting job...\n")
        touchjob.run()

        print("Job ID      : %s" % (touchjob.id))

        # List all jobs that are known by the adaptor.
        # This should show our job as well.
        print("\nListing active jobs: ")
        for job in js.list():
            print(" * %s" % job)

        # wait for our job to complete
        print("\n...waiting for job...\n")
        touchjob.wait()

        print("Job State   : %s" % (touchjob.state))
        print("Exitcode    : %s" % (touchjob.exit_code))
        print("Exec. hosts : %s" % (touchjob.execution_hosts))
        print("Create time : %s" % (touchjob.created))
        print("Start time  : %s" % (touchjob.started))
        print("End time    : %s" % (touchjob.finished))

        js.close()
        return 0

    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (ex.type, (str(ex))))
        # Get the whole traceback in case of an exception -
        # this can be helpful for debugging the problem
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        return -1

if __name__ == "__main__":
    sys.exit(main())




__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import re
import os
import sys
import errno
import tempfile

import radical.utils              as ru

from  .  import misc              as sumisc
from  .  import pty_shell_factory as supsf
from  .  import pty_process       as supp
from  .. import session           as ss
from  .. import filesystem        as sfs

from  .  import pty_exceptions    as ptye

from  ..import exceptions         as rse


# ------------------------------------------------------------------------------
#
_PTY_TIMEOUT = 2.0


# ------------------------------------------------------------------------------
#
# iomode flags
#
IGNORE   = 0    # discard stdout / stderr
MERGED   = 1    # merge stdout and stderr
SEPARATE = 2    # fetch stdout and stderr individually (one more hop)
STDOUT   = 3    # fetch stdout only, discard stderr
STDERR   = 4    # fetch stderr only, discard stdout


# ------------------------------------------------------------------------------
#
DEFAULT_PROMPT = "[\$#%>\]]\s*$"


# --------------------------------------------------------------------
#
class PTYShell (object) :
    """
    This class wraps a shell process and runs it as a :class:`PTYProcess`.  The
    user of this class can start that shell, and run arbitrary commands on it.

    The shell to be run is expected to be POSIX compliant (bash, dash, sh, ksh
    etc.) -- in particular, we expect the following features:
    ``$?``,
    ``$!``,
    ``$#``,
    ``$*``,
    ``$@``,
    ``$$``,
    ``$PPID``,
    ``>&``,
    ``>>``,
    ``>``,
    ``<``,
    ``|``,
    ``||``,
    ``()``,
    ``&``,
    ``&&``,
    ``wait``,
    ``kill``,
    ``nohup``,
    ``shift``,
    ``export``,
    ``PS1``, and
    ``PS2``.

    Note that ``PTYShell`` will change the shell prompts (``PS1`` and ``PS2``),
    to simplify output parsing.  ``PS2`` will be empty, ``PS1`` will be set
    ``PROMPT-$?->`` -- that way, the prompt will report the exit value of the
    last command, saving an extra roundtrip.  Users of this class should be
    careful when setting other prompts -- see :func:`set_prompt` for more
    details.

    Usage Example::

        # start the shell, find its prompt.
        self.shell = saga.utils.pty_shell.PTYShell("ssh://user@rem.host.net/",
                                                   contexts, self._logger)

        # run a simple shell command, merge stderr with stdout.  $$ is the pid
        # of the shell instance.
        ret, out, _ = self.shell.run_sync (" mkdir -p /tmp/data.$$/" )

        # check if mkdir reported success
        if  ret != 0 :
            raise saga.NoSuccess ("failed to prepare dir (%s)(%s)" % (ret, out))

        # stage some data from a local string variable
        # into a file on the remote system
        self.shell.stage_to_remote (src = pbs_job_script,
                                    tgt = "/tmp/data.$$/job_1.pbs")

        # check size of staged script (this is actually done on PTYShell level
        # already, with no extra hop):
        ret, out, _ = self.shell.run_sync("stat -c '%s' /tmp/data.$$/job_1.pbs")
        if  ret != 0 :
            raise saga.NoSuccess ("failed to check size (%s)(%s)" % (ret, out))

        assert (len(pbs_job_script) == int(out))


    **Data Staging and Data Management:**


    The PTYShell class does not only support command execution, but also basic
    data management: for SSH based shells, it will create a tunneled scp/sftp
    connection for file staging.  Other data management operations (mkdir, size,
    list, ...) are executed either as shell commands, or on the scp/sftp channel
    (if possible on the data channel, to keep the shell pty free for concurrent
    command execution).  Ssh tunneling is implemented via ssh.v2 'ControlMaster'
    capabilities (see `ssh_config(5)`).

    For local shells, PTYShell will create an additional shell pty for data
    management operations.


    **Asynchronous Notifications:**

    A third pty process will be created for asynchronous notifications.  For
    that purpose, the shell started on the first channel will create a named
    pipe, at::

      $HOME/.radical/saga/adaptors/shell/async.$$

    ``$$`` here represents the pid of the shell process.  It will also set the
    environment variable ``SAGA_ASYNC_PIPE`` to point to that named pipe -- any
    application running on the remote host can write event messages to that
    pipe, which will be available on the local end (see below).  `PTYShell`
    leaves it unspecified what format those messages have, but messages are
    expected to be separated by newlines.

    An adaptor using `PTYShell` can subscribe for messages via::

      self.pty_shell.subscribe (callback)

    where callback is a Python callable.  PTYShell will listen on the event
    channel *in a separate thread* and invoke that callback on any received
    message, passing the message text (sans newline) to the callback.

    An example usage: the command channel may run the following command line::

      ( sh -c 'sleep 100 && echo "job $$ done" > $SAGA_ASYNC_PIPE" \
                         || echo "job $$ fail" > $SAGA_ASYNC_PIPE" ) &

    which will return immediately, and send a notification message at job
    completion.

    Note that writes to named pipes are not atomic.  From POSIX:

    ``A write is atomic if the whole amount written in one operation is not
    interleaved with data from any other process. This is useful when there are
    multiple writers sending data to a single reader. Applications need to know
    how large a write request can be expected to be performed atomically. This
    maximum is called {PIPE_BUF}. This volume of IEEE Std 1003.1-2001 does not
    say whether write requests for more than {PIPE_BUF} bytes are atomic, but
    requires that writes of {PIPE_BUF} or fewer bytes shall be atomic.`

    Thus the user is responsible for ensuring that either messages are smaller
    than *PIPE_BUF* bytes on the remote system (usually at least 1024, on Linux
    usually 4096), or to lock the pipe on larger writes.


    **Automated Restart, Timeouts:**

    For timeout and restart semantics, please see the documentation to the
    underlying :class:`saga.utils.pty_process.PTYProcess` class.

    """

    # TODO:
    #   - on client shell activitites, also mark the master as active, to
    #     avoid timeout garbage collection.
    #   - use ssh mechanisms for master timeout (and persist), as custom
    #     mechanisms will interfere with gc_timout.

    # unique ID per connection, for debugging
    _pty_id = 0

    # ----------------------------------------------------------------
    #
    def __init__ (self, url, session=None, logger=None, cfg=None, posix=True,
            interactive=True):

        if logger : self.logger  = logger
        else      : self.logger  = ru.Logger('radical.saga.pty')

        if session: self.session = session
        else      : self.session = ss.Session(default=True)

        self.logger.debug ("PTYShell init %s" % self)

        self.url         = url          # describes the shell to run
        self.posix       = posix        # /bin/sh compatible?
        self.interactive = interactive  # bash -i ?
        self.latency     = 0.0          # set by factory
        self.cp_slave    = None         # file copy channel

        self.initialized = False

        self.pty_id       = PTYShell._pty_id
        PTYShell._pty_id += 1

        name = None
        if isinstance(cfg, str):
            name = cfg
            cfg  = None
        self.cfg = ru.Config('radical.saga.session', name=name, cfg=cfg)
        self.cfg = self.cfg.pty

        # get prompt pattern from config, or use default
        self.prompt    = self.cfg.get('prompt_pattern', DEFAULT_PROMPT)
        self.prompt_re = re.compile ("^(.*?)%s"    % self.prompt, re.DOTALL)
        self.logger.info ("PTY prompt pattern: %s" % self.prompt)

        # we need a local dir for file staging caches.  At this point we use
        # $HOME, but should make this configurable (FIXME)
        self.base = os.environ['HOME'] + '/.radical/saga/adaptors/shell/'

        try:
            os.makedirs (self.base)

        except OSError as e:
            if e.errno == errno.EEXIST and os.path.isdir (self.base):
                pass
            else:
                raise rse.NoSuccess ("could not create staging dir: %s" % e) \
                      from e


        self.factory    = supsf.PTYShellFactory   ()
        self.pty_info   = self.factory.initialize (self.url,    self.session,
                                                   self.prompt, self.logger,
                                                   self.cfg,    self.posix,
                                                   interactive=self.interactive)
        self.pty_shell  = self.factory.run_shell  (self.pty_info)

        self._trace ('init : %s' % self.pty_shell.command)

        self.initialize ()


    # ----------------------------------------------------------------
    #
    def _trace (self, msg) :

      # print " === %5d : %s : %s" % (self._pty_id, self.pty_shell, msg)
      # self.logger.debug(" === %5d : %s : %s", self._pty_id,self.pty_shell,msg)
        pass


    # ----------------------------------------------------------------
    #
    def __del__ (self) :

        self.finalize(kill_pty=True)


    # ----------------------------------------------------------------
    #
    def initialize (self) :
        """ initialize the shell connection.  """

        with self.pty_shell.rlock :

            if  self.initialized :
                self.logger.warn ("initialization race")
                return


            if  self.posix :
                # run a POSIX compatible shell, usually /bin/sh, in interactive
                # mode also, turn off tty echo
                command_shell = "exec /bin/sh -i"

                # use custom shell if so requested
                if  self.cfg.get('shell'):
                    command_shell = "exec %s" % self.cfg['shell']
                    self.logger.info("custom command shell: %s" % command_shell)


                self.logger.debug("running command shell: %s" % command_shell)
                self.pty_shell.write(" stty -echo ; %s\n" % command_shell)

                # make sure this worked, and that we find the prompt. We use
                # a versatile prompt pattern to account for the custom shell
                # case.
                _, out = self.find ([self.prompt])

                # make sure this worked, and that we find the prompt. We use
                # a versatile prompt pattern to account for the custom shell
                # case.
                try :
                    # set and register new prompt
                    self.run_async(" set HISTFILE=$HOME/.saga_history;"
                                   " PS1='PROMPT-$?->';"
                                   " PS2='';"
                                   " PROMPT_COMMAND='';"
                                   " export PS1 PS2 PROMPT_COMMAND 2>&1 >/dev/null;"
                                   " cd $HOME 2>&1 >/dev/null\n")
                    self.set_prompt (new_prompt="PROMPT-(\d+)->$")

                    self.logger.debug ("got new shell prompt")

                except Exception as e :
                    raise rse.NoSuccess ("Shell on target host failed: %s" % e)\
                          from e

            # got a command shell, finally!
            self.pty_shell.flush ()
            self.initialized = True
            self.finalized   = False


    # ----------------------------------------------------------------
    #
    def finalize (self, kill_pty=False) :

        try :
            if  kill_pty and self.pty_shell :
                with self.pty_shell.rlock :
                    if not self.finalized :
                        self.pty_shell.finalize ()
                        self.finalized = True

        except Exception:
            pass


    # ----------------------------------------------------------------
    #
    def alive (self, recover=False) :
        """
        The shell is assumed to be alive if the shell processes lives.
        Attempt to restart shell if recover==True
        """

        with self.pty_shell.rlock :

            try :
                return self.pty_shell.alive (recover)

            except Exception as e :
                raise ptye.translate_exception (e) from e


    # ----------------------------------------------------------------
    #
    def find_prompt (self, timeout=_PTY_TIMEOUT) :
        """
        If run_async was called, a command is running on the shell.  find_prompt
        can be used to collect its output up to the point where the shell prompt
        re-appears (i.e. when the command finishes).


        Note that this method blocks until the command finishes.  Future
        versions of this call may add a timeout parameter.
        """

        with self.pty_shell.rlock :

            try :

                match = None
                fret  = None

                while fret is None :
                    fret, match = self.pty_shell.find ([self.prompt], timeout)

              # self.logger.debug("find prompt '%s' in '%s'"
              #                  % (self.prompt, match))
                ret, txt = self._eval_prompt (match)

                return (ret, txt)

            except Exception as e :
                raise ptye.translate_exception (e) from e


    # ----------------------------------------------------------------
    #
    def find (self, patterns, timeout=-1) :
        """
        Note that this method blocks until pattern is found in the shell I/O.
        """

        with self.pty_shell.rlock :

            try :
                return self.pty_shell.find (patterns, timeout=timeout)

            except Exception as e :
                raise ptye.translate_exception (e) from e


    # ----------------------------------------------------------------
    #
    def set_prompt (self, new_prompt) :
        """
        :type  new_prompt:  string
        :param new_prompt:  a regular expression matching the shell prompt

        The new_prompt regex is expected to be a regular expression with one set
        of catching brackets, which MUST return the previous command's exit
        status.  This method will send a newline to the client, and expects to
        find the prompt with the exit value '0'.

        As a side effect, this method will discard all previous data on the pty,
        thus effectively flushing the pty output.

        By encoding the exit value in the command prompt, we safe one roundtrip.
        The prompt on Posix compliant shells can be set, for example, via::

          PS1='PROMPT-$?->'; export PS1

        The newline in the example above allows to nicely anchor the regular
        expression, which would look like::

          PROMPT-(\d+)->$

        The regex is compiled with 're.DOTALL', so the dot character matches
        all characters, including line breaks.  Be careful not to match more
        than the exact prompt -- otherwise, a prompt search will swallow stdout
        data.  For example, the following regex::

          PROMPT-(.+)->$

        would capture arbitrary strings, and would thus match *all* of::

          PROMPT-0->ls
          data/ info
          PROMPT-0->

        and thus swallow the ls output...

        Note that the string match *before* the prompt regex is non-gready -- if
        the output contains multiple occurrences of the prompt, only the match
        up to the first occurence is returned.
        """

        def escape (txt) :
            pat = re.compile(r'\x1b[^m]*m')
            return pat.sub ('', txt)


        with self.pty_shell.rlock :

            old_prompt     = self.prompt
            self.prompt    = new_prompt
            self.prompt_re = re.compile("^(.*?)%s\s*$" % self.prompt, re.DOTALL)

            retries  = 0
            triggers = 0

            while True :

                try :
                    # make sure we have a non-zero waiting delay (default to
                    # 1 second)
                    delay = 10 * self.latency
                    if  not delay :
                        delay = 1.0

                    # FIXME: how do we know that _PTY_TIMOUT suffices?  In
                    #        particular if we actually need to flush...
                    fret, match = self.pty_shell.find ([self.prompt], delay)

                    if  fret is None :

                        retries += 1
                        if  retries > 10 :
                            self.prompt = old_prompt
                            raise rse.BadParameter("Cannot use new prompt,"
                                               "parsing failed (10 retries)")

                        self.pty_shell.write ("\n")
                        self.logger.debug("sent prompt trigger again (%d)"
                                         % retries)
                        triggers += 1
                        continue


                    # found a match -- lets see if this is working now...
                    ret, _ = self._eval_prompt (match)

                    if  ret != 0 :
                        self.prompt = old_prompt
                        raise rse.BadParameter ("could not parse exit value (%s)"
                                            % match)

                    # prompt looks valid...
                    break

                except Exception as e :
                    self.prompt = old_prompt
                    raise ptye.translate_exception (e, "Could not set shell prompt")\
                          from e


            # got a valid prompt -- but we have to sync the output again in
            # those cases where we had to use triggers to actually get the
            # prompt
            if triggers > 0 :
                self.run_async (' printf "SYNCHRONIZE_PROMPT\n"')

                # FIXME: better timout value?
                fret, match = self.pty_shell.find(["SYNCHRONIZE_PROMPT"],
                                                  timeout=10.0)

                if  fret is None :
                    # not find prompt after blocking?  BAD!  Restart the shell
                    self.finalize (kill_pty=True)
                    raise rse.NoSuccess ("Could not synchronize prompt detection")

                self.find_prompt ()



    # ----------------------------------------------------------------
    #
    def _eval_prompt (self, data, new_prompt=None) :
        """
        This method will match the given data against the current prompt regex,
        and expects to find an integer as match -- which is then returned, along
        with all leading data, in a tuple
        """

        with self.pty_shell.rlock :

            try :

                prompt    = self.prompt
                prompt_re = self.prompt_re

                if  new_prompt :
                    prompt    = new_prompt
                    prompt_re = re.compile ("^(.*)%s\s*$" % prompt, re.DOTALL)


                result = None
                if  not data :
                    raise rse.NoSuccess("cannot not parse prompt (%s), data: %s"
                                     % (prompt, data))

                result = prompt_re.match (data)

                if  not result :
                    raise rse.NoSuccess("could not parse prompt (%s) (%s)"
                                   % (prompt, data))

                txt = result.group (1)
                ret = 0

                if  len (result.groups ()) != 2 :
                    if  new_prompt :
                        self.logger.warn("prompt captures no exit code (%s)"
                                         % prompt)
                      # raise NoSuccess ("prompt captures no exit code (%s)"
                      #                 % prompt)

                else :
                    try :
                        ret = int(result.group (2))
                    except ValueError :
                        # apparently, this is not an integer. Print a warning,
                        # and assume success -- the calling entity needs to
                        # evaluate the remainder...
                        ret = 0
                        self.logger.warn("prompt unusable for error checks (%s)"
                                         % prompt)
                        txt += "\n%s" % result.group (2)

                # if that worked, we can permanently set new_prompt
                if  new_prompt :
                    self.set_prompt (new_prompt)

                return (ret, txt)

            except Exception as e :

                raise ptye.translate_exception (e, "Could not eval prompt") \
                      from e




    # ----------------------------------------------------------------
    #
    def run_sync (self, command, iomode=None, new_prompt=None) :
        """
        Run a shell command, and report exit code, stdout and stderr (all three
        will be returned in a tuple).  The call will block until the command
        finishes (more exactly, until we find the prompt again on the shell's
        I/O stream), and cannot be interrupted.

        :type  command: string
        :param command: shell command to run.

        :type  iomode:  enum
        :param iomode:  Defines how stdout and stderr are captured.

        :type  new_prompt:  string
        :param new_prompt:  regular expression matching the prompt after
        command succeeded.

        We expect the ``command`` to not to do stdio redirection, as this is we want
        to capture that separately.  We *do* allow pipes and stdin/stdout
        redirection.  Note that SEPARATE mode will break if the job is run in
        the background


        The following iomode values are valid:

          * *IGNORE:*   both stdout and stderr are discarded, `None` will be
                        returned for each.
          * *MERGED:*   both streams will be merged and returned as stdout;
                        stderr will be `None`.  This is the default.
          * *SEPARATE:* stdout and stderr will be captured separately, and
                        returned individually.  Note that this will require
                        at least one more network hop!
          * *STDOUT:*   only stdout is captured, stderr will be `None`.
          * *STDERR:*   only stderr is captured, stdout will be `None`.
          * *None:*     do not perform any redirection -- this is effectively
                        the same as `MERGED`

        If any of the requested output streams does not return any data, an
        empty string is returned.


        If the command to be run changes the prompt to be expected for the
        shell, the ``new_prompt`` parameter MUST contain a regex to match the
        new prompt.  The same conventions as for set_prompt() hold -- i.e. we
        expect the prompt regex to capture the exit status of the process.
        """

        with self.pty_shell.rlock :

            self._trace ("run sync  : %s" % command)
            self.pty_shell.flush ()

            # we expect the shell to be in 'ground state' when running
            # a syncronous command -- thus we can check if the shell is alive
            # before doing so, and restart if needed
            if not self.pty_shell.alive (recover=True) :
                raise rse.IncorrectState ("Can't run command -- shell died:\n%s"
                                      % self.pty_shell.autopsy ())

            try :

                command = command.strip ()
                if command.endswith ('&') :
                    raise rse.BadParameter("run_sync can only run foreground jobs"
                                       "('%s')" % command)

                redir = ""
                _err  = "/tmp/radical.saga.ssh-job.stderr.$$"

                if iomode is None    : redir  =  ""
                if iomode == IGNORE  : redir  =  " 1>>/dev/null 2>>/dev/null"
                if iomode == MERGED  : redir  =  " 2>&1"
                if iomode == SEPARATE: redir  =  " 2>%s" % _err
                if iomode == STDOUT  : redir  =  " 2>/dev/null"
                if iomode == STDERR  : redir  =  " 2>&1 1>/dev/null"

                self.logger.debug    ('run_sync: %s%s'   % (command, redir))
                self.pty_shell.write (          "%s%s\n" % (command, redir))


                # If given, switch to new prompt pattern right now...
                prompt = self.prompt
                if  new_prompt :
                    prompt = new_prompt

                # command has been started - now find prompt again.
                fret, match = self.pty_shell.find ([prompt], timeout=-1.0)

                if  fret is None :
                    # not find prompt after blocking?  BAD!  Restart the shell
                    self.finalize (kill_pty=True)
                    raise rse.IncorrectState (
                            "run_sync failed, no prompt (%s)" % command)


                ret, txt = self._eval_prompt (match, new_prompt)

                stdout = txt
                stderr = None

                if  iomode in [SEPARATE, STDERR]:

                    self.pty_shell.write(" cat %s\n" % _err)
                    fret, match = self.pty_shell.find ([self.prompt],
                                                       timeout=-1.0)  # blocks
                    if  fret is None :
                        # not find prompt after blocking?  BAD!  Restart shell
                        self.finalize (kill_pty=True)
                        raise rse.IncorrectState (
                                "run_sync failed, no prompt (%s)" % command)

                    _ret, _stderr = self._eval_prompt (match)

                    if  _ret :
                        raise rse.IncorrectState(
                                "run_sync failed, no stderr (%s: %s)"
                                % (_ret, _stderr))

                    stderr =  _stderr

                if  iomode == STDERR :
                    # got stderr in branch above
                    stdout = None

                elif iomode == IGNORE:
                    stdout = None
                    stderr = None

                return (ret, stdout, stderr)

            except Exception as e :
                raise ptye.translate_exception (e) from e


    # ----------------------------------------------------------------
    #
    def run_async (self, command) :
        """
        Run a shell command, but don't wait for prompt -- just return.  It is up
        to caller to eventually search for the prompt again (see
        :func:`find_prompt`.  Meanwhile, the caller can interact with the called
        command, via the I/O channels.

        :type  command: string
        :param command: shell command to run.

        For async execution, we don't care if the command is doing i/o
        redirection or not.
        """

        with self.pty_shell.rlock :

            self._trace ("run async : %s" % command)
            self.pty_shell.flush ()

            # we expect the shell to be in 'ground state' when running an
            # asyncronous command -- thus we can check if the shell is alive
            # before doing so, and restart if needed
            if not self.pty_shell.alive (recover=True) :
                raise rse.IncorrectState ("Cannot run command:\n%s"
                                      % self.pty_shell.autopsy ())

            try :
                command = command.strip ()
                self.send ("%s\n" % command)

            except Exception as e :
                raise ptye.translate_exception (e) from e


    # ----------------------------------------------------------------
    #
    def send (self, data) :
        """
        send data to the shell.  No newline is appended!
        """

        with self.pty_shell.rlock :

            if not self.pty_shell.alive (recover=False) :
                raise rse.IncorrectState("Cannot send data:\n%s"
                                       % self.pty_shell.autopsy ())

            try :
                self.pty_shell.write ("%s" % data)

            except Exception as e :
                raise ptye.translate_exception (e) from e


    # ----------------------------------------------------------------
    #
    def write_to_remote (self, src, tgt) :
        """
        :type  src: string
        :param src: data to be staged into the target file

        :type  tgt: string
        :param tgt: path to target file to staged to
                    The tgt path is not an URL, but expected to be a path
                    relative to the shell's URL.

        The content of the given string is pasted into a file (specified by tgt)
        on the remote system.  If that file exists, it is overwritten.
        A NoSuccess exception is raised if writing the file was not possible
        (missing permissions, incorrect path, etc.).
        """

        try :

          # self._trace ("write     : %s -> %s" % (src, tgt))

            # FIXME: make this relative to the shell's pwd?  Needs pwd in
            # prompt, and updating pwd state on every find_prompt.

            # first, write data into a tmp file
            fhandle, fname = tempfile.mkstemp(suffix='.tmp',
                                              prefix='rs_pty_staging_')
            os.write(fhandle, str.encode(src))
            os.fsync(fhandle)
            os.close(fhandle)

            ret = self.stage_to_remote (fname, tgt)

            os.remove (fname)

            return ret

        except Exception as e :
            raise ptye.translate_exception (e) from e


    # ----------------------------------------------------------------
    #
    def read_from_remote (self, src) :
        """
        :type  src: string
        :param src: path to source file to staged from
                    The src path is not an URL, but expected to be a path
                    relative to the shell's URL.
        """

        try :

          # self._trace ("read      : %s" % src)

            # FIXME: make this relative to the shell's pwd?  Needs pwd in
            # prompt, and updating pwd state on every find_prompt.

            # first, write data into a tmp file
            fhandle, fname = tempfile.mkstemp(suffix='.tmp', prefix='rs_pty_staging_')
            _ = self.stage_from_remote (src, fname)
            os.close(fhandle)

            os.system('sync')  # WTF?  Why do I need this?

            fhandle2 = open(fname, 'r')
            out      = fhandle2.read()
            fhandle2.close()

            os.remove(fname)

            return out

        except Exception as e :
            raise ptye.translate_exception (e) from e


    # ----------------------------------------------------------------
    #
    def stage_to_remote (self, src, tgt, cp_flags=None) :
        """
        :type  src: string
        :param src: path of local source file to be stage from.
                    The tgt path is not an URL, but expected to be a path
                    relative to the current working directory.

        :type  tgt: string
        :param tgt: path to target file to stage to.
                    The tgt path is not an URL, but expected to be a path
                    relative to the shell's URL.
        """

        self._trace ("stage to  : %s -> %s" % (src, tgt))

        # FIXME: make this relative to the shell's pwd?  Needs pwd in
        # prompt, and updating pwd state on every find_prompt.

        try :
            return self.run_copy_to (src, tgt, cp_flags)

        except Exception as e :
            raise ptye.translate_exception (e) from e


    # ----------------------------------------------------------------
    #
    def stage_from_remote (self, src, tgt, cp_flags="") :
        """
        :type  src: string
        :param src: path to source file to stage from.
                    The tgt path is not an URL, but expected to be a path
                    relative to the shell's URL.

        :type  tgt: string
        :param tgt: path of local target file to stage to.
                    The tgt path is not an URL, but expected to be a path
                    relative to the current working directory.
        """

        self._trace ("stage from: %s -> %s" % (src, tgt))

        # FIXME: make this relative to the shell's pwd?  Needs pwd in
        # prompt, and updating pwd state on every find_prompt.

        try :
            return self.run_copy_from (src, tgt, cp_flags)

        except Exception as e :
            raise ptye.translate_exception (e) from e


    # --------------------------------------------------------------------------
    #
    def run_copy_to (self, src, tgt, cp_flags=None) :
        """
        This initiates a slave copy connection.   Src is interpreted as local
        path, tgt as path on the remote host.

        Now, this is ugly when over sftp: sftp supports recursive copy, and
        wildcards, all right -- but for recursive copies, it wants the target
        dir to exist -- so, we have to check if the local src is a  dir, and if
        so, we first create the target before the copy.  Worse, for wildcards we
        have to do a local expansion, and then to do the same for each entry...
        """

        if cp_flags is None:
            cp_flags = ''

        with self.pty_shell.rlock :

            self._trace ("copy  to  : %s -> %s" % (src, tgt))
            self.pty_shell.flush ()

            info = self.pty_info
            repl = dict (list({'src'      : src,
                               'tgt'      : tgt,
                               'cp_flags' : cp_flags
                          }.items ()) + list(info.items ()))

            # at this point, we do have a valid, living master
            s_cmd = info['scripts'][info['copy_mode']]['copy_to']    % repl
            s_in  = info['scripts'][info['copy_mode']]['copy_to_in'] % repl
            posix = info['scripts'][info['copy_mode']]['copy_is_posix']

            if  not s_in :
                # this code path does not use an interactive shell for copy --
                # so the above s_cmd is all we want to run, really.  We get
                # do not use the chached cp_slave in this case, but just run the
                # command.  We do not have a list of transferred files though,
                # yet -- that should be parsed from the proc output.

                cp_proc = supp.PTYProcess (s_cmd, cfg=self.cfg)
                out = cp_proc.wait ()
                if  cp_proc.exit_code :
                    raise ptye.translate_exception(rse.NoSuccess(
                                             "file copy failed: %s" % out))

                return list()


            # this code path uses an interactive shell to transfer files, of
            # some form, such as sftp.  Get the shell cp_slave from cache, and
            # run the actual copy command.
            if  not self.cp_slave :
                self._trace ("get cp slave")
                self.cp_slave = self.factory.get_cp_slave (s_cmd, info, posix)

            self.cp_slave.flush ()
            if  'sftp' in s_cmd :
                # prepare target dirs for recursive copy, if needed
                import glob
                src_list = glob.glob (src)
                for s in src_list :
                    if  os.path.isdir (s) :
                        prep = "mkdir %s/%s\n" % (tgt, os.path.basename (s))
                        # TODO: handle multiple levels of creation

                        self.cp_slave.flush()
                        self.cp_slave.write("%s\n" % prep)
                        self.cp_slave.find(['[\$\>\]]\s*$'], -1)
                        # TODO: check return values

                if cp_flags == sfs.CREATE_PARENTS and os.path.split(tgt)[0]:
                    # TODO: this needs to be numeric and checking the flag
                    prep = "mkdir %s\n" % os.path.dirname(tgt)
                    # TODO: this doesn't deal with multiple levels of creation

                    self.cp_slave.flush()
                    self.cp_slave.write("%s\n" % prep)
                    self.cp_slave.find(['[\$\>\]]\s*$'], -1)
                    # TODO: check return values

            self.cp_slave.flush()
            _ = self.cp_slave.write("%s\n" % s_in)
            _, out = self.cp_slave.find(['[\$\>\]]\s*$'], -1)

            # FIXME: we don't really get exit codes from copy
            # if  self.cp_slave.exit_code != 0 :
            #     raise rse.NoSuccess._log (info['logger'],
            #                              "file copy failed: %s" % str(out))

            if 'Invalid flag' in out :
                raise rse.NoSuccess._log(info['logger'],
                                    "unsupported sftp version %s" % str(out))
            if 'No such file or directory' in out :
                raise rse.DoesNotExist._log(info['logger'],
                                           "file copy failed: %s" % str(out))

            if 'is not a directory' in out :
                raise rse.BadParameter._log(info['logger'],
                                           "File copy failed: %s" % str(out))

            if  'sftp' in s_cmd :
                if 'not found' in out :
                    raise rse.BadParameter._log(info['logger'],
                                               "file copy failed: %s" % out)


            # we interpret the first word on the line as name of src file -- we
            # will return a list of those
            lines = out.split ('\n')
            files = []

            for line in lines :

                elems = line.split (' ', 2)

                if  elems :

                    f = elems[0]

                    # remove quotes
                    if  f :
                        if f[ 0] in ["'", '"', '`']: f = f[1:  ]
                        if f[-1] in ["'", '"', '`']: f = f[ :-1]

                    # ignore empty lines
                    if  f :

                        files.append (f)

            info['logger'].debug ("copy done: %s" % files)

            return files


    # --------------------------------------------------------------------------
    #
    def run_copy_from (self, src, tgt, cp_flags="") :
        """
        This initiates a slave copy connection.   Src is interpreted as path on
        the remote host, tgt as local path.

        We have to do the same mkdir trick as for the run_copy_to, but here we
        need to expand wildcards on the *remote* side :/
        """

        with self.pty_shell.rlock :

            self._trace ("copy  from: %s -> %s" % (src, tgt))
            self.pty_shell.flush ()

            info = self.pty_info
            repl = dict (list({'src'      : src,
                          'tgt'      : tgt,
                          'cp_flags' : cp_flags}.items()) + list(info.items ()))

            # at this point, we do have a valid, living master
            s_cmd = info['scripts'][info['copy_mode']]['copy_from']    % repl
            s_in  = info['scripts'][info['copy_mode']]['copy_from_in'] % repl
            posix = info['scripts'][info['copy_mode']]['copy_is_posix']

            if  not s_in :
                # this code path does not use an interactive shell for copy --
                # so the above s_cmd is all we want to run, really.  We get
                # do not use the chached cp_slave in this case, but just run the
                # command.  We do not have a list of transferred files though,
                # yet -- that should be parsed from the proc output.
                cp_proc = supp.PTYProcess (s_cmd, cfg=self.cfg)
                cp_proc.wait ()
                if  cp_proc.exit_code :
                    raise ptye.translate_exception(rse.NoSuccess(
                        "file copy failed: exit code %s" % cp_proc.exit_code))

                return list()

            if  not self.cp_slave :
                self._trace ("get cp slave")
                self.cp_slave = self.factory.get_cp_slave (s_cmd, info, posix)

            self.cp_slave.flush ()
            prep = ""

            if  'sftp' in s_cmd :
                # prepare target dirs for recursive copy, if needed
                self.cp_slave.write (" ls %s\n" % src)
                _, out = self.cp_slave.find (["^sftp> "], -1)

                src_list = out[1].split('\n')

                for s in src_list :
                    if  os.path.isdir (s) :
                        prep += "lmkdir %s/%s\n" % (tgt, os.path.basename (s))

            self.cp_slave.flush ()
            _      = self.cp_slave.write("%s%s\n" % (prep, s_in))
            _, out = self.cp_slave.find (['[\$\>\]] *$'], -1)

            # FIXME: we don't really get exit codes from copy
          # if  self.cp_slave.exit_code != 0 :
          #     raise NoSuccess._log (info['logger'],
          #                           "file copy failed: %s" % out)

            if 'Invalid flag' in out :
                raise rse.NoSuccess._log(info['logger'],
                                        "sftp version not supported (%s)" % out)

            if 'No such file or directory' in out :
                raise rse.DoesNotExist._log(info['logger'],
                                           "file copy failed: %s" % out)

            if 'is not a directory' in out :
                raise rse.BadParameter._log(info['logger'],
                                           "file copy failed: %s" % out)

            if  'sftp' in s_cmd :
                if 'not found' in out :
                    raise rse.BadParameter._log(info['logger'],
                                               "file copy failed: %s" % out)


            # we run copy with -v, so get a list of files which have been copied
            # -- we parse that list and return it.  we interpret the *second*
            # word on the line as name of src file.
            lines = out.split ('\n')
            files = []

            for line in lines :

                elems = line.split (' ', 3)

                if  elems and len(elems) > 1 and elems[0] == 'Fetching' :

                    f = elems[1]

                    # remove quotes
                    if f:
                        if f[ 0] in ["'", '"', '`']: f = f[1:  ]
                        if f[-1] in ["'", '"', '`']: f = f[ :-1]

                    # ignore empty lines
                    if  f :
                        files.append (f)

            info['logger'].debug ("copy done: %s" % files)

            return files


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "MIT"

import redis

r = redis.Redis (host='localhost')

keep = False

print("-----------------------------------------------")

keys = r.keys ("*")
keys.sort()

for k in keys :
    t = str(r.type (k), 'utf-8')

    if  t == 'hash' :
        d = r.hgetall (k)
        if not 'url' in d :
            if not keep: r.delete (k)
        print("%-35s [%-6s] %s " % (k, t, d))
        if not keep : r.delete (k)

    elif t == 'list' :
        print("%-35s [%-6s] %s " % (k, t, r.lrange (k, 0, -1)))
        if not keep: r.delete (k)

    elif t == 'set' :
        print("%-35s [%-6s] %s " % (k, t, r.smembers (k)))
        if not keep: r.delete (k)

    else :
        print("%-35s [%-6s] %s " % (k, t, r.get (k)))
        if not keep: r.delete (k)


print("-----------------------------------------------")





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils               as ru
import radical.utils.signatures as rus

from .constants  import *
from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base       as sab
from ..namespace import entry      as nsentry
from .. import attributes          as sa
from .. import session             as ss
from .. import task                as st


# ------------------------------------------------------------------------------
#
class Entry (nsentry.Entry, sa.Attributes) :


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry', 
                  rus.optional ((ru.Url, str)), 
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session), 
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, url=None, flags=READ, session=None, 
                  _adaptor=None, _adaptor_state={}, _ttype=None) : 
        '''
        url:       saga.Url
        flags:     flags enum
        session:   saga.Session
        ret:       obj
        '''

        # param checks
        url = ru.Url (url)

        self._nsentry = super  (Entry, self)
        self._nsentry.__init__ (url, flags, session, 
                                _adaptor, _adaptor_state, _ttype=_ttype)


        # set attribute interface properties
        self._attributes_allow_private (True)
        self._attributes_camelcasing   (True)
        self._attributes_extensible    (True, getter=self._attribute_getter, 
                                              setter=self._attribute_setter,
                                              lister=self._attribute_lister,
                                              caller=self._attribute_caller)

        # register properties with the attribute interface 
        self._attributes_register   (ATTRIBUTE, None, sa.STRING, sa.SCALAR, sa.READONLY)
        self._attributes_register   (OBJECT,    None, sa.ANY,    sa.SCALAR, sa.READONLY)
        self._attributes_register   (TTL,       None, sa.INT,    sa.SCALAR, sa.WRITEABLE)

        self._attributes_set_setter (TTL,    self.set_ttl)
        self._attributes_set_getter (TTL,    self.get_ttl)

        self._attributes_set_setter (OBJECT, self.store_object)
        self._attributes_set_getter (OBJECT, self.retrieve_object)



    # --------------------------------------------------------------------------
    #
    @classmethod
    @rus.takes   ('Entry', 
                  rus.optional ((ru.Url, str)), 
                  rus.optional (int, rus.nothing), 
                  rus.optional (ss.Session), 
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (st.Task)
    def create (cls, url=None, flags=READ, session=None, ttype=None) :
        '''
        url:       saga.Url
        flags:     saga.advert.flags enum
        session:   saga.Session
        ttype:     saga.task.type enum
        ret:       saga.Task
        '''

        if not flags : flags = 0
        _nsentry = super (Entry, cls)
        return _nsentry.create (url, flags, session, ttype=ttype)



    # --------------------------------------------------------------------------
    #
    # attribute methods
    #
    # NOTE: we do not yet pass ttype, as async calls are not yet supported by
    # the attribute interface
    #
    @rus.takes   ('Entry', 
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.anything, st.Task))
    def _attribute_getter (self, key, ttype=None) :

        return self._adaptor.attribute_getter (key)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry', 
                  str,
                  rus.anything,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def _attribute_setter (self, key, val, ttype=None) :

        return self._adaptor.attribute_setter (key, val)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry', 
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (rus.anything), st.Task))
    def _attribute_lister (self, ttype=None) :

        return self._adaptor.attribute_lister ()


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry', 
                  str, 
                  int, 
                  callable, 
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.anything, st.Task))
    def _attribute_caller (self, key, id, cb, ttype=None) :

        return self._adaptor.attribute_caller (key, id, cb)



    # --------------------------------------------------------------------------
    #
    # advert methods
    #
    @rus.takes   ('Entry', 
                  float, 
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def set_ttl  (self, ttl=-1.0, ttype=None) : 
        """
        ttl :           int
        ttype:          saga.task.type enum
        ret:            None / saga.Task
        """

        return self._adaptor.set_ttl (ttl, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry', 
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((float, st.Task))
    def get_ttl  (self, ttype=None) : 
        """
        ttype:          saga.task.type enum
        ret:            int / saga.Task
        """

        return self._adaptor.get_ttl (ttype=ttype)



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry', 
                  object,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def store_object (self, object, ttype=None) : 
        """
        object :        <object type>
        ttype:          saga.task.type enum
        ret:            None / saga.Task
        """
        return self._adaptor.store_object (object, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry', 
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((object, st.Task))
    def retrieve_object (self, ttype=None) : 
        """
        ttype:          saga.task.type enum
        ret:            any / saga.Task
        """
        return self._adaptor.retrieve_object (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry', 
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def delete_object (self, ttype=None) : 
        """
        ttype:          saga.task.type enum
        ret:            None / saga.Task
        """
        return self._adaptor.delete_object (ttype=ttype)






__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .constants import *
from .file      import File
from .directory import Directory








__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


''' Provides a parser class for the file transfer specification as it is
    defined in GFD.90, sction 4.1.3.
'''

from ... import exceptions as se


# 4.1.3 File Transfer Specifications (GFD90 p 176-177)
#
# The syntax of a file transfer directive for the job description is modeled on
# the LSF syntax (LSF stands for Load Sharing Facility, a commercial job
# scheduler by Platform Computing), and has the general syntax:
# local_file operator remote_file
# Both the local_file and the remote_file can be URLs. If they are not URLs,
# but full or relative pathnames, then the local_file is relative to the host
# where the submission is executed, and the remote_file is evaluated on the
# execution host of the job. The operator is one of the following four:
#
# '>'  copies the local file to the remote file before the job starts.
#      Overwrites the remote file if it exists.
# '>>' copies the local file to the remote file before the job starts.
#      Appends to the remote file if it exists.
# '<'  copies the remote file to the local file after the job finishes.
#      Overwrites the local file if it exists.
# '<<' copies the remote file to the local file after the job finishes.
#      Appends to the local file if it exists.


# ------------------------------------------------------------------------------
#
class TransferDirectives(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, directives=None):

        self._in_overwrite  = list()
        self._in_append     = list()
        self._out_overwrite = list()
        self._out_append    = list()

        if not directives:
            directives = []

        for d in directives:

            if (d.count('>') > 2) or (d.count('<') > 2):
                msg = "'%s' is not a valid transfer d string."
                raise se.BadParameter(msg)

            elif '>>' in d:
                (loc, rem) = d.split('>>')
                self._in_append.append([loc.strip(), rem.strip()])

            elif '>' in d:
                (loc, rem) = d.split('>')
                self._in_overwrite.append([loc.strip(), rem.strip()])

            elif '<<' in d:
                (loc, rem) = d.split('<<')
                self._out_append.append([loc.strip(), rem.strip()])

            elif '<' in d:
                (loc, rem) = d.split('<')
                self._out_overwrite.append([loc.strip(), rem.strip()])

            else:
                msg = "'%s' is not a valid transfer directive string." % d
                raise se.BadParameter(msg)


    # --------------------------------------------------------------------------
    #
    def _to_string_list(self):

        slist = list()

        for (loc, rem) in self._in_overwrite:
            slist.append('%s > %s' % (loc, rem))

        for (loc, rem) in self._in_append:
            slist.append('%s >> %s' % (loc, rem))

        for (loc, rem) in self._out_overwrite:
            slist.append('%s < %s' % (loc, rem))

        for (loc, rem) in self._out_append:
            slist.append('%s << %s' % (loc, rem))

        return slist


    # --------------------------------------------------------------------------
    #
    def __str__(self):

        return str(self._to_string_list())


    # --------------------------------------------------------------------------
    #
    @property
    def in_overwrite(self):

        return self._in_overwrite


    # --------------------------------------------------------------------------
    #
    @property
    def in_append(self):

        return self._in_append


    # --------------------------------------------------------------------------
    #
    @property
    def out_overwrite(self):

        return self._out_overwrite


    # --------------------------------------------------------------------------
    #
    @property
    def out_append(self):

        return self._out_append


    # --------------------------------------------------------------------------
    #
    @property
    def string_list(self):

        return self._to_string_list()


# ------------------------------------------------------------------------------
#
def _test_():

    tdp = TransferDirectives(["ab","a>c", "c>>d","f<a","g<<h"])

    print(tdp.in_append)
    print(tdp.in_overwrite)
    print(tdp.out_append)
    print(tdp.out_overwrite)


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" This is an example which shows how to access Amazon EC2 clouds via the SAGA
    resource package.

    In order to run this example, you need to set the following environment
    variables:

    * EC2_ACCESS_KEY:     your Amazon EC2 ID
    * EC2_SECRET_KEY:     your Amazon EC2 KEY
    * EC2_SSH_KEYPAIR_ID: name of ssh keypair within EC2
    * EC2_SSH_KEYPAIR:    your ssh keypair to use to access the VM, e.g.,
                          /home/username/.ssh/id_rsa_ec2
"""


import os
import sys
import time

import radical.saga as rs


# ------------------------------------------------------------------------------
#
def main():

    # In order to connect to EC2, we need an EC2 ID and KEY. We read those
    # from the environment.
    ec2_ctx = rs.Context('EC2')
    ec2_ctx.user_id  = os.environ['EC2_ACCESS_KEY']
    ec2_ctx.user_key = os.environ['EC2_SECRET_KEY']

    # The SSH keypair we want to use the access the EC2 VM. If the keypair is
    # not yet registered on EC2 saga will register it automatically.  This
    # context specifies the key for VM startup, ie. the VM will be configured to
    # accept this key
    ec2keypair_ctx = rs.Context('EC2_KEYPAIR')
    ec2keypair_ctx.token    = os.environ['EC2_KEYPAIR_ID']
    ec2keypair_ctx.user_key = os.environ['EC2_KEYPAIR']
    ec2keypair_ctx.user_id  = 'root'  # the user id on the target VM

    # We specify the *same* ssh key for ssh access to the VM.  That now should
    # work if the VM go configured correctly per the 'EC2_KEYPAIR' context
    # above.
    ssh_ctx = rs.Context('SSH')
    ssh_ctx.user_id  = 'root'
    ssh_ctx.user_key = os.environ['EC2_KEYPAIR']

    session = rs.Session(False)  # FALSE: don't use other (default) contexts
    session.contexts.append(ec2_ctx)
    session.contexts.append(ec2keypair_ctx)
    session.contexts.append(ssh_ctx)

    cr  = None  # compute resource handle
    rid = None  # compute resource ID
    try:

        # ----------------------------------------------------------------------
        #
        # reconnect to VM (ID given in ARGV[1])
        #
        if len(sys.argv) > 1:
            
            rid = sys.argv[1]

            # reconnect to the given resource
            print('reconnecting to %s' % rid)
            cr = rs.resource.Compute(id=rid, session=session)
            print('reconnected  to %s' % rid)
            print("  state : %s (%s)" % (cr.state, cr.state_detail))


        # ----------------------------------------------------------------------
        #
        # start a new VM
        #
        else:

            # start a VM if needed
            # in our session, connect to the EC2 resource manager
            rm = rs.resource.Manager("ec2://aws.amazon.com/", session=session)

            # Create a resource description with an image and an OS template,.
            # We pick a small VM and a plain Ubuntu image...
            cd = rs.resource.ComputeDescription()
            cd.image    = 'ami-0256b16b'    # plain ubuntu
            cd.template = 'Small Instance'

            # Create a VM instance from that description.
            cr  = rm.acquire(cd)
            rid = cr.id

            print("\nWaiting for VM to become active...")


        # ----------------------------------------------------------------------
        #
        # use the VM
        #
        # Wait for the VM to 'boot up', i.e., become 'ACTIVE'
        cr.wait(rs.resource.ACTIVE)

        # Query some information about the newly created VM
        print("Created VM: %s"      %  cr.id)
        print("  state   : %s (%s)" % (cr.state, cr.state_detail))
        print("  access  : %s"      %  cr.access)

        # give the VM some time to start up comlpetely, otherwise the subsequent
        # job submission might end up failing...
        time.sleep(60)

        # create a job service which uses the VM's access URL (cr.access)
        js = rs.job.Service(cr.access, session=session)

        jd = rs.job.Description()
        jd.executable = '/bin/sleep'
        jd.arguments = ['30']

        job = js.create_job(jd)
        job.run()

        print("\nRunning Job: %s" % job.id)
        print("  state : %s" % job.state)
        job.wait()
        print("  state : %s" % job.state)


    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (ex.type, (str(ex))))
        raise


    except Exception as e:
        # Catch all other exceptions
        print("An Exception occured: %s " % e)
        raise


    finally:

        # ----------------------------------------------------------------------
        #
        # shut VM down (only when id was specified on command line)
        if cr and rid:
            cr.destroy()
            print("\nDestroyed VM: %s" % cr.id)
            print("  state : %s (%s)" % (cr.state, cr.state_detail))


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":
    
    sys.exit(main())





#!/usr/bin/env python

__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" This examples shows how to run a job on a remote Condor gateway
    using the 'condor' job adaptor.
"""

import sys
import getpass

import radical.saga as rs


URL = "condor+ssh://gw68.quarry.iu.teragrid.org?WhenToTransferOutput=ON_EXIT&should_transfer_files=YES&notification=Always",
URL = "condor+gsissh://login.osgconnect.net"
URL = "condor+gsissh://xd-login.opensciencegrid.org"
URL = "condor+gsissh://submit-1.osg.xsede.org"

def main():
    try:
        # Your ssh identity on the remote machine.
      # ctx = rs.Context("ssh")
      # ctx.user_id = getpass.getuser()  # Change if necessary

        session = rs.Session()
      # session.add_context(ctx)

        # create a job service for the local machine. both, 'fork' and
        # 'local' schemes trigger the local job adaptor.
        js = rs.job.Service(URL, session=session)

        # describe our job
        jd = rs.job.Description()

        # environment, executable & arguments. We use '/bin/sleep' to simulate
        # a job that runs for $RUNTIME seconds.
        jd.name            = 'testjob'
        jd.project         = 'TG-MCB090174'
        jd.environment     = {'RUNTIME': '0'}
        jd.wall_time_limit = 2 # minutes

        jd.executable = '/bin/sleep'
        jd.arguments = ["$RUNTIME"]

        jd.output          = "saga_condorjob.stdout"
        jd.error           = "saga_condorjob.stderr"
 
 #       jd.candidate_hosts = ["FNAL_FERMIGRID", "cinvestav", "SPRACE",
 #                             "NYSGRID_CORNELL_NYS1", "Purdue-Steele",
 #                             "MIT_CMS_CE2", "SWT2_CPB", "AGLT2_CE_2",
 #                             "UTA_SWT2", "GridUNESP_CENTRAL",
 #                             "USCMS-FNAL-WC1-CE3"]

        # create the job (state: New)
        sleepjob = js.create_job(jd)

        # check our job's id and state
        print("Job ID    : %s" % (sleepjob.id))
        print("Job State : %s" % (sleepjob.state))

        print("\n...starting job...\n")
        sleepjob.run()

        print("Job ID    : %s" % (sleepjob.id))
        print("Job State : %s" % (sleepjob.state))

        print("\nListing active jobs: ")
        for job in js.list():
            print(" * %s" % job)

        # disconnect / reconnect
        sleepjob_clone = js.get_job(sleepjob.id)

        # wait for our job to complete
        print("\n...waiting for job...\n")
        sleepjob_clone.wait()

        print("Job State   : %s" % (sleepjob_clone.state))
        print("Exitcode    : %s" % (sleepjob_clone.exit_code))
        print("Exec. hosts : %s" % (sleepjob_clone.execution_hosts))
        print("Create time : %s" % (sleepjob_clone.created))
        print("Start time  : %s" % (sleepjob_clone.started))
        print("End time    : %s" % (sleepjob_clone.finished))

        js.close()
        return 0

    except rs.SagaException as ex:
        print("An exception occured: %s " % ((str(ex))))
        # get the whole traceback in case of an exception -
        # this can be helpful for debugging the problem
        print(" *** %s" % ex.traceback)
        return -1

if __name__ == "__main__":
    sys.exit(main())




__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils               as ru
import radical.utils.signatures    as rus

from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base       as sab

from .. import sasync
from .. import task                as st
from .. import base                as sb
from .. import session             as ss
from .. import exceptions          as se
from .. import constants           as sc

from .  import description         as descr
from .  import resource            as resrc
from .  import constants           as c


# ------------------------------------------------------------------------------
#
class Manager (sb.Base, sasync.Async) :
    """
    In the context of RADICAL-SAGA, a *ResourceManager* is a service which
    asserts control over a set of resources.  That manager can, on request,
    render control over subsets of those resources (resource slices) to an
    application.

    This :class:`Manager` class represents the contact point to such
    ResourceManager instances -- the application can thus acquire compute, data
    or network resources, according to some resource specification, for a bound
    or unbound amount of time.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Manager',
                  rus.optional (str, ru.Url),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, url=None, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :
        """
        __init__(url)

        Create a new Manager instance. Connect to a remote resource management endpoint.

        :type  url: :class:`saga.Url`
        :param url: resource management endpoint
        """

        # param checks
        _url   = ru.Url(url)
        scheme = _url.scheme.lower()

        if not session :
            session = ss.Session (default=True)

        self._base = super  (Manager, self)
        self._base.__init__ (scheme, _adaptor, _adaptor_state,
                             _url, session, ttype=_ttype)


    # --------------------------------------------------------------------------
    #
    @classmethod
    @rus.takes   ('Manager',
                  rus.optional ((ru.Url, str)),
                  rus.optional (ss.Session),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (st.Task)
    def create   (cls, url_in=None, session=None, ttype=sc.SYNC) :
        """
        This is the asynchronous class constructor, returning
        a :class:`saga:Task` instance.  For details on the accepted parameters,
        please see the description of :func:`__init__`.
        """

        return cls (url_in, session, _ttype=ttype)._init_task


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Manager',
                  rus.optional (rus.one_of (c.COMPUTE, c.STORAGE, c.NETWORK)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (str), st.Task))
    def list     (self, rtype=None, ttype=None) :
        """
        list(rtype=None)

        List known resource instances (which can be acquired).
        Returns a list of IDs.

        :type  rtype: None or enum (COMPUTE | STORAGE | NETWORK)
        :param rtype: filter for one or more resource types

        """
        return self._adaptor.list (rtype, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Manager',
                  rus.optional (str),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((descr.Description, st.Task))
    def get_description (self, rid, ttype=None) :
        """
        get_description(rid)

        Get the resource :class:`Description` for the specified resource.

        :type  rid: str
        :param rid: identifies the resource to be described.
        """

        # TODO / NOTE: if rid is None, should we return a description of
        # the managed resources?
        return self._adaptor.get_description (id, ttype=ttype)

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Manager',
                  rus.optional (rus.one_of (c.COMPUTE, c.STORAGE, c.NETWORK)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (str), st.Task))
    def list_templates (self, rtype=None, ttype=None) :
        """
        list_templates(rtype=None)

        List template names available for the specified resource type(s).
        Returns a list of strings.

        :type  rtype: None or enum (COMPUTE | STORAGE | NETWORK)
        :param rtype: filter for one or more resource types

        """
        return self._adaptor.list_templates (rtype, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Manager',
                  rus.optional (str),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((descr.Description, st.Task))
    def get_template (self, name, ttype=None) :
        """
        get_template(name)

        Get a :class:`Description` for the specified template.

        :type  name: str
        :param name: specifies the name of the template

        The returned resource description instance may not have all attributes
        filled, and may in fact not sufficiently complete to allow for
        successful resource acquisition.  The only guaranteed attribute in the
        returned description is `TEMPLATE`, containing the very template id
        specified in the call parameters.
        """

        return self._adaptor.get_template (name, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Manager',
                  rus.optional (rus.one_of (c.COMPUTE, c.STORAGE, c.NETWORK)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (str), st.Task))
    def list_images (self, rtype=None, ttype=None) :
        """
        list_images(rtype=None)

        List image names available for the specified resource type(s).
        Returns a list of strings.

        :type  rtype: None or enum (COMPUTE | STORAGE | NETWORK)
        :param rtype: filter for one or more resource types
        """
        return self._adaptor.list_images (rtype, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Manager',
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((dict, st.Task))
    def get_image (self, name, ttype=None) :
        """
        get_image(name)

        Get a description string for the specified image.

        :type  name: str
        :param name: specifies the image name
        """

        return self._adaptor.get_image (name, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Manager',
                  (str, ru.Url, descr.Description),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((resrc.Resource, st.Task))
    def acquire  (self, spec, ttype=None) :
        """
        acquire(desc)

        Create a new :class:`saga.resource.Resource` handle for a
        resource specified by the description.

        :type  spec: :class:`Description` or Url
        :param spec: specifies the resource

        Depending on the `RTYPE` attribute in the description, the
        returned resource may be a :class:`saga.resource.Compute`,
        :class:`saga.resource.Storage` or :class:`saga.resource.Network`
        instance.

        If the `spec` parameter is
        """

        if  isinstance (spec, ru.Url) or \
            isinstance (spec, str)  :

            return self._adaptor.acquire_by_id(spec, ttype=ttype)

        else :

            # make sure at least 'executable' is defined
            if spec.rtype is None:
                raise se.BadParameter ("resource type undefined in description")

            # spec_copy = descr.Description ()
            # spec._attributes_deep_copy (spec_copy)

            return self._adaptor.acquire (spec, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Manager',
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def destroy  (self, rid, ttype=None) :
        """
        destroy(rid)

        Destroy / release a resource.

        :type  rid   : string
        :param rid   : identifies the resource to be released
        """

        return self._adaptor.destroy (rid, ttype=ttype)

  # FIXME: add
  # templates = property(list_templates, get_template)     # dict{string: Description}
  # images    = property(list_images,    get_image)        # dict{string: dict}
  # resources = property(list,           get_description)  # dict{string: Description}


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import threading as mt

import radical.utils         as ru
import radical.utils.logger  as rul

from   ..exceptions import *


# ------------------------------------------------------------------------------
# adaptor base class
#
class Base(object, metaclass=ru.Singleton):
    '''
    adaptor base class

    We only need one instance of this adaptor per process (actually per
    engine, but engine is a singleton, too...) -- the engine will though
    create new CPI implementation instances as needed (one per SAGA API
    object).
    '''

    # --------------------------------------------------------------------------
    #
    # FIXME: phase out adaptor_options
    #
    def __init__(self, adaptor_info, adaptor_options=None, expand_env=True):

        # FIXME: engine is loading cfg already, here we load again...

        self._info    = adaptor_info
        self._name    = adaptor_info['name']
        self._schemas = adaptor_info['schemas']

        self._lock    = mt.RLock()
        self._logger  = ru.Logger('radical.saga.api')

        self._cfg     = ru.Config(module='radical.saga.adaptors',
                                  name=self._name,
                                  expand=expand_env)

        if 'enabled' not in self._cfg:
            self._cfg['enabled'] = True


    # --------------------------------------------------------------------------
    #
    # if sanity_check() is commented out here, then we will only load adaptors
    # which implement the method themselves.
    #
    def sanity_check(self):
        '''
        This method can be overloaded by adaptors to check runtime conditions on
        adaptor load time.  The adaptor should raise an exception if it will not
        be able to function properly in the given environment, e.g. due to
        missing dependencies etc.
        '''

        pass


    # --------------------------------------------------------------------------
    #
    def register(self):
        '''
        Adaptor registration function. The engine calls this during startup
        to retrieve the adaptor information.
        '''

        return self._info


    # --------------------------------------------------------------------------
    #
    def get_name(self):

        return self._name


    # --------------------------------------------------------------------------
    #
    def get_schemas(self):

        return self._schemas


    # --------------------------------------------------------------------------
    #
    def get_info(self):

        return self._info


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


# assert (False), 'deprecated'

import radical.utils.testing  as rut
import radical.saga           as rs


# ----------------------------------------------------------------------
#
def configure_jd (cfg, jd) :

    for a in ['job_walltime_limit' ,
              'job_project'        ,
              'job_queue'          ,
              'job_total_cpu_count',
              'job_spmd_variation' ] :

        if  cfg.get (a, None) :
            jd.set_attribute (a, cfg[a])

    return jd


# ----------------------------------------------------------------------
#
def assert_exception (cfg, e) :

    ni = cfg.get ('not_implemented', 'warn')

    if  'NotImplemented' in str(e) and ni == 'warn' :
        print("WARNING: %s")
        return

    else :
        assert (False), "unexpected exception '%s'" % e
        raise e


# ----------------------------------------------------------------------
#
def silent_cancel (obj) :

    if  not isinstance (obj, list) :
        obj = [obj]

    for o in obj :
        try :
            o.cancel ()
        except Exception :
            pass


# ----------------------------------------------------------------------
#
def silent_close (obj) :

    if  not isinstance (obj, list) :
        obj = [obj]

    for o in obj :
        try :
            o.close ()
        except Exception :
            pass


# ------------------------------------------------------------------------------
#
class TestConfig (rut.TestConfig):

    # --------------------------------------------------------------------------
    #
    def __init__ (self, cfg_file):

        # initialize configuration.  We only use the 'rs.tests' category from
        # the config file.
        rut.TestConfig.__init__ (self, cfg_file, 'radical.saga.tests')

        # setup a saga session for the tests
        # don't populate session with default contexts...
        self.session = rs.Session (default=False)

        # attempt to create a context from the test config
        if  self.context_type :

            c = rs.Context (self.context_type)

            c.user_id    = self.context_user_id
            c.user_pass  = self.context_user_pass
            c.user_cert  = self.context_user_cert
            c.user_proxy = self.context_user_proxy

            # add it to the session
            self.session.add_context (c)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils.signatures as rus

from .. import attributes       as sa
from .. import exceptions       as se

from .  import constants        as c


# ------------------------------------------------------------------------------
#
class Description (sa.Attributes) :
    """ 
    The resource description class. 

    Resource descriptions are used for two purposes:

      * an application can pass a description instances to a
        :class:`saga.resource.Manager` instance, to request control 
        over the resource slice described in the description; 

      * an application can request a resource's description for 
        inspection of resource properties.

    There are three specific types of descriptions: 

      * :class:`saga.resource.ComputeDescription` for the description of
        resources with compute capabilities;

      * :class:`saga.resource.StorageDescription` for the description of
        resources with data storage capabilities;

      * :class:`saga.resource.NetworkDescription` for the description of
        resources with communication capabilities.

    There is at this point no notion of resources which combine different
    capabilities.

    For all these capabilities, the following attributes are supported:

      * `RType`      : `Enum`, describing the capabilities of the resource
                       (`COMPUTE`, `STORAGE` or `NETWORK`)

      * `Template`   : `String`, a backend specific resource class with some
                       pre-defined hardware properties to apply to the resource.

      * `Image`      : `String`, a backend specific resource class with some 
                       pre-defined software properties to apply to the resource.

      * `Dynamic`    : `Boolean, if `True` signifies that the resource may
                       dynamically change its properties at runtime

      * `Start`      : `Integer (seconds) since epoch when the resource is 
                       expected to enter / when the resource entered `ACTIVE` 
                       state.

      * `End`        : `Integer (seconds) since epoch when the resource is 
                       expected to enter / when the resource entered a `FINAL` 
                       state.

      * `Duration`   : `Integer`, seconds for which the resource is expected to
                       remain / the resource remained in `ACTIVE` state.

      * `MachineOS`  : `String`, for `COMPUTE` resources, specifies the
                       operating system type running on that resource.

      * `MachineArch : `String`, for `COMPUTE` resources, specifies the
                       machine architecture of that resource.

      * `Size`       : `Integer`, for `COMPUTE` resources, specifies the
                       number of process slots provided, for `STORAGE` resource
                       specifies the number of bytes, of the resource.

      * `Memory`     : `Integer`, for `COMPUTE` resources, specifies the
                       number of bytes provided as memory.

      * `Access`     : `String`, usually an URL, which specifies the contact
                       point for the resource capability interface / service
                       interface.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Description', 
                  rus.optional (dict))
    @rus.returns (rus.nothing)
    def __init__ (self, d=None):
        """
        __init__()

        Create a new Description instance.
        """

        # set attribute interface properties

        self._attributes_extensible  (False)
        self._attributes_camelcasing (True)

        # register properties with the attribute interface

        self._attributes_register(c.RTYPE       , None , sa.ENUM  , sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.TEMPLATE    , None , sa.STRING, sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.IMAGE       , None , sa.STRING, sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.DYNAMIC     , False, sa.BOOL  , sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.START       , None , sa.TIME  , sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.END         , None , sa.TIME  , sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.DURATION    , None , sa.TIME  , sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.MACHINE_OS  , None , sa.STRING, sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.MACHINE_ARCH, None , sa.STRING, sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.SIZE        , 1    , sa.INT   , sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.MEMORY      , None , sa.STRING, sa.SCALAR, sa.WRITEABLE) 
        self._attributes_register(c.ACCESS      , None , sa.STRING, sa.SCALAR, sa.WRITEABLE) 

        self._attributes_set_enums(c.RTYPE, [c.COMPUTE ,
                                             c.STORAGE ,
                                             c.NETWORK ])

        # FIXME: initialization should be done in Attributes: initialization
        # from dict or from other attributable
        #
        if  d :
            for key in d.list_attributes () :
                self.set_attribute (key, d.get_attribute (key)) 


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Description', 
                  ('Description', dict))
    @rus.returns ('Description')
    def __deepcopy__ (self, memo) :
        """
        An alias for `clone()`.
        """
        return self.clone()


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Description', 
                  rus.optional ('Description'))
    @rus.returns ('Description')
    def clone (self, other=None) :
        """ 
        clone()

        Implements deep copy. 

        Unlike the default python assignment (copy object reference),
        a deep copy will create a new object instance with the same state --
        after a deep copy, a change on one instance will not affect the other.
        """

        # a job description only has attributes - so create a new instance,
        # clone the attribs, and done.
        if not other :
            other = Description ()

        return self._attributes_deep_copy (other)


# ------------------------------------------------------------------------------
#
class ComputeDescription (Description) : 
    """
    A `ComputeDescription` is a specific description for a resource which
    provides compute capabilities, i.e. the ability to run application
    executables / code.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('ComputeDescription', 
                  rus.optional (dict))
    @rus.returns (rus.nothing)
    def __init__ (self, d=None) :

        if  d :
            if  c.RTYPE in d and d[c.RTYPE] != c.COMPUTE :
                raise se.BadParameter ("Cannot create ComputeResource type '%s'"
                                    % d[c.RTYPE])

        self._descr = super  (ComputeDescription, self)
        self._descr.__init__ (d)

        self.rtype = c.COMPUTE


# ------------------------------------------------------------------------------
#
class StorageDescription (Description) :
    """
    A `StorageDescription` is a specific description for a resource which
    provides storage capabilities, i.e. the ability to persistently store,
    organize and retrieve data files.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('StorageDescription', 
                  rus.optional (dict))
    @rus.returns (rus.nothing)
    def __init__ (self, d=None) :

        if  d :
            if  c.RTYPE in d and d[c.RTYPE] != c.STORAGE :
                raise se.BadParameter ("Cannot create StorageResource type '%s'"
                                    % d[c.RTYPE])

        self._descr = super  (StorageDescription, self)
        self._descr.__init__ (d)

        self.rtype = c.STORAGE


# ------------------------------------------------------------------------------
#
class NetworkDescription (Description) :
    """
    A `NetworkDescription` is a specific description for a resource which
    provides network capabilities.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('NetworkDescription', 
                  rus.optional (dict))
    @rus.returns (rus.nothing)
    def __init__ (self, d=None) :

        if  d:
            if  c.RTYPE in d and d[c.RTYPE] != c.NETWORK :
                raise se.BadParameter ("Cannot create NetworkResource type '%s'"
                                    % d[c.RTYPE])

        self._descr = super  (NetworkDescription, self)
        self._descr.__init__ ()

        self.rtype = c.NETWORK


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .. import task as st


# ------------------------------------------------------------------------------
#
# 'forward' declaration of job.Container
#
class Container (st.Container):
    """ :todo: document me

        .. py:attribute:: jobs

           The (read-only) `jobs` property returns a list of all job objects in the container.

           :rtype: :class:`saga.job.Job` list

        .. py:attribute:: size

           The (read-only) `size` property returns the number of job objectis in the container.

           :rtype: int

        .. py:attribute:: states

           The (read-only) `states` property returns a list of states that represent the 
           states of the individual jobs in the container. 

           :rtype: list
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self) :

        self._task_container = super  (Container, self)
        self._task_container.__init__ ()

        import radical.saga.attributes as sa

        self._attributes_register   ("Jobs",   [], sa.ANY, sa.VECTOR, sa.READONLY)
        self._attributes_set_getter ("Jobs",   self.get_jobs)


    # --------------------------------------------------------------------------
    #
    def get_jobs (self) :
        """ This is similar to get_tasks(), but returns only Job typed entries
        from the container.
        """

        tasks = self.get_tasks ()
        jobs  = []

        from . import job as sjob

        for task in tasks :
            if isinstance (task, sjob.Job) :
                jobs.append (task)

        return jobs







#!/usr/bin/env python

__author__    = "Andre Merzky, Matteo Turilli, Ole Weidner"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "MIT"


import os
import sys
import time

import radical.saga as rs


"""  
This is an example which shows how to access Amazon EC2 clouds via the SAGA
resource package. The code expects the environment variables EC2_ACCESS_KEY and
EC2_SECRET_KEY to contain the respective authentication tokens required for EC2
access.  It also expects EC2_KEYPAIR to point to the ssh key to be used in the
EC2 keypair authentication.

This program has different modes of operation:

  * *Help* ::
  
    # python examples/resource/ec2.py -h
    
    Usage:

        %s -l             :  list    VMs
        %s -c <id> [...]  :  create  VM
        %s -u <id> [...]  :  use     VMs (run jobs)
        %s -d <id> [...]  :  destroy VMs

    Environment:
       
       EC2_URL        : backend manager service endpoint
       EC2_ACCESS_KEY : id  for backend access
       EC2_SECRET_KEY : key for backend access
       EC2_KEYPAIR_ID : name of keypair for VM access
       EC2_KEYPAIR    : public ssh key for VM access


  * *Listing* of templates and existing VMs on EC2::

    # python examples/resource/ec2.py -l
    
    compute resources
      [ec2://aws.amazon.com/]-[i-cba515ab]
      [ec2://aws.amazon.com/]-[i-f93f2299]
    
    compute templates
      Micro Instance
      Small Instance
      Medium Instance
      Large Instance
      Extra Large Instance
      High-Memory Extra Large Instance
      High-Memory Double Extra Large Instance
      High-Memory Quadruple Extra Large Instance
      Extra Large Instance
      Double Extra Large Instance
      High-CPU Medium Instance
      High-CPU Extra Large Instance
      Cluster Compute Quadruple Extra Large Instance
      Cluster Compute Eight Extra Large Instance
      Cluster GPU Quadruple Extra Large Instance
      High Memory Cluster Eight Extra Large
      High Storage Eight Extra Large Instance
    

  * *Creating* a VM instance on EC2::

    # python examples/resource/ec2.py -c

    Created VM
      id           : 
      state        : PENDING (pending)
      access       : None


  * *Using* a VM instance on EC2::
  
    # python examples/resource/ec2.py -u '[ec2://aws.amazon.com/]-[i-e0d2ad8a]'

    connecting to [ec2://aws.amazon.com/]-[i-e0d2ad8a]
      id           : [ec2://aws.amazon.com/]-[i-e0d2ad8a]
      state        : PENDING (pending)
      wait for ACTIVE state
      state        : ACTIVE (running)
    running job
      job state    : Running
      job state    : Done


  * *Destroying* a VMs instance on EC2::

    # python examples/resource/ec2.py -d '[ec2://aws.amazon.com/]-[i-e0d2ad8a]'

    reconnecting to id [ec2://aws.amazon.com/]-[i-e0d2ad8a]
      id           : [ec2://aws.amazon.com/]-[i-e0d2ad8a]
      state        : ACTIVE (running)
      access       : ssh://107.21.154.248/
    
    shutting down [ec2://aws.amazon.com/]-[i-e0d2ad8a] (ACTIVE)
      state        : EXPIRED (destroyed by user)
    
"""


# ------------------------------------------------------------------------------
#
# helper
#
def usage (msg = None) :

    if  msg :
        print("\n    Error: %s\n"  %  msg)

    print("""
    Usage:

        %s -l             :  list    VMs
        %s -c <id> [...]  :  create  VM
        %s -u <id> [...]  :  use     VMs (run jobs)
        %s -d <id> [...]  :  destroy VMs


    Environment:
       
       EC2_URL        : backend manager service endpoint
       EC2_ACCESS_KEY : id  for backend access
       EC2_SECRET_KEY : key for backend access
       EC2_KEYPAIR_ID : name of keypair for VM access
       EC2_KEYPAIR    : public ssh key for VM access


    """)

    if msg : sys.exit (-1)
    else   : sys.exit ( 0)


# ------------------------------------------------------------------------------
#
def state2str (state) :

    if state == rs.resource.UNKNOWN  : return "UNKNOWN"
    if state == rs.resource.NEW      : return "NEW   "
    if state == rs.resource.PENDING  : return "PENDING"
    if state == rs.resource.ACTIVE   : return "ACTIVE"
    if state == rs.resource.CANCELED : return "CANCELED"
    if state == rs.resource.EXPIRED  : return "EXPIRED"
    if state == rs.resource.DONE     : return "DONE  "
    if state == rs.resource.FAILED   : return "FAILED"



# ------------------------------------------------------------------------------
#
# set up the connection to EC2
#

if not 'EC2_URL'        in os.environ : usage ("no %s in environment" % 'EC2_URL'       )
if not 'EC2_ACCESS_KEY' in os.environ : usage ("no %s in environment" % 'EC2_ACCESS_KEY')
if not 'EC2_SECRET_KEY' in os.environ : usage ("no %s in environment" % 'EC2_SECRET_KEY')
if not 'EC2_KEYPAIR_ID' in os.environ : usage ("no %s in environment" % 'EC2_KEYPAIR_ID')
if not 'EC2_KEYPAIR'    in os.environ : usage ("no %s in environment" % 'EC2_KEYPAIR'   )

server = rs.Url(os.environ['EC2_URL'])


# in order to connect to EC2, we need an EC2 ID and KEY
c1 = rs.Context ('ec2')
c1.user_id  = os.environ['EC2_ACCESS_KEY']
c1.user_key = os.environ['EC2_SECRET_KEY']
c1.server   = server

# in order to access a created VM, we additionally need to point to the ssh
# key which is used for EC2 VM contextualization, i.e. as EC2 'keypair'.
# If the keypair is not yet registered on EC2, it will be registered by SAGA
# -- but then a user_key *must* be specified (only the public key is ever
# transfererd to EC2).
c2 = rs.Context ('ec2_keypair')
c2.token     = os.environ['EC2_KEYPAIR_ID']
c2.user_cert = os.environ['EC2_KEYPAIR']
c2.user_id   = 'ubuntu'         # the user id on the target VM
c2.server    = server

# we create a session for all SAGA interactions, and attach the respective
# security contexts.  Those are now avail for all SAGA objects created in
# that session
s = rs.Session (False)  # FALSE: don't use any other (default) contexts
s.contexts.append (c1)
s.contexts.append (c2)

# in this session, connect to the EC2 resource manager
rm  = rs.resource.Manager (server, session=s)


# --------------------------------------------------------------------------
# 
# setup is done, evaluate command line parameters
#
args = sys.argv[1:]

# --------------------------------------------------------------------------
if  '-l' in args :

    args.remove ('-l')
    if  len (args) > 0 :
        usage ("no additional args allowed on '-l'")

    # list known VMs (compute resources)
    print("\ncompute resources")
    for cr_id in rm.list () :
        print("  %s" % cr_id)

    # list the available VM templates
    print("\ncompute templates")
    for tmp in rm.list_templates () :
        print("  %s" % tmp)

    # we can also list the available OS images, as per below -- but since
    # the list of OS images avaialble on EC2 is *huge*, this operation is
    # rather slow (libcloud does one additional hop per image, for
    # inspection)

    # {'name': 'None (cube-1-0-5-2012-09-07)', 'ispublic': 'true', 'state': 'available', 'rootdevicetype': 'instance-store', 'imagetype': 'machine'}
    print("\nOS images")
    
    descr = None
    ispublic = None

    for osi in rm.list_images () :
        descr = rm.get_image (osi)

        if descr['ispublic'] == 'true' :
            ispublic = 'public'
        else:
            ispublic = 'private'

        print("  %s - %s, %s, %s" % (osi, descr['name'], ispublic,
                                     descr['state']))

    print()
    sys.exit (0)


# --------------------------------------------------------------------------
elif  '-c' in args :

    args.remove ('-c')
    if  len (args) == 0 :
        usage ("additional args required on '-c'")

    for image in args :

        print("\ncreating an instance from image %s" % image)

        # create a resource description with an image and an OS template, out of
        # the ones listed above.  We pick a small VM and a plain Ubuntu image...
        cd = rs.resource.ComputeDescription ()
        cd.image    = image
        cd.template = 'Small Instance'

        # create a VM instance with that description, and inspect it for some
        # detailes
        cr = rm.acquire (cd)
        print("\nCreated VM")
        print("  id           : %s"       %  cr.id)
        print("  state        : %s (%s)"  %  (state2str(cr.state), cr.state_detail))
        print("  access       : %s"       %  cr.access)

    sys.exit (0)

# --------------------------------------------------------------------------
elif  '-u' in args :

    args.remove ('-u')
    if  len (args) == 0 :
        usage ("additional args required on '-u'")

    # we want to reconnect to running VMs, specified by their IDs
    for vm_id in args :

        print("\nconnecting to %s" % vm_id)

        # get a handle on that VM, and print some information
        cr = rm.acquire (vm_id)

        print("  id           : %s"       %  cr.id)
        print("  state        : %s (%s)"  %  (state2str(cr.state), cr.state_detail))

        # make sure the machine is not in final state already
        if  cr.state in [rs.resource.EXPIRED,
                         rs.resource.DONE,
                         rs.resource.FAILED] :
            print("  VM %s is alrady in final state"  %  vm_id)
            continue

        # we only can run jobs on ACTIVE machines -- so lets wait until the VM
        # is in that state.
        # Note: the careful coder will spot the subtle race condition between the 
        # check above and the check on this line... ;-)
        if cr.state != rs.resource.ACTIVE :
          print("  wait for ACTIVE state")
          cr.wait (rs.resource.ACTIVE)

          # Once a VM comes active, it still needs to boot and setup the ssh
          # daemon to be usable -- we thus wait for a while
          time.sleep (60)

        print("  state        : %s (%s)"  %  (state2str(cr.state), cr.state_detail))
        print("  access       : %s"       %   cr.access)


        # The session created above contains the ssh context to access the VM
        # instance -- that context was created from the ec2_keypair context
        # which was earlier used for VM contextualization.  So we use that
        # session to create a job service instance for that VM:
        js = rs.job.Service (cr.access, session=s)

        print("running job")
        # all ready: do the deed!
        j = js.run_job ('sleep 10')
        print("  job state    : %s"  %  j.state)
        j.wait ()
        print("  job state    : %s"  %  j.state)

    print()
    sys.exit (0)



# --------------------------------------------------------------------------
elif  '-d' in args :

    args.remove ('-d')
    if  len (args) == 0 :
        usage ("additional args required on '-d'")

    # we want to reconnect to running VMs, specified by their IDs
    for vm_id in args :

        print("\nreconnecting to id %s" % vm_id)

        # get a handle on that VM, and print some information
        cr = rm.acquire (vm_id)

        print("  id           : %s"       %  cr.id)
        print("  state        : %s (%s)"  %  (state2str(cr.state), cr.state_detail))
        print("  access       : %s"       %  cr.access)

        if cr.state in [rs.resource.EXPIRED,
                        rs.resource.DONE,
                        rs.resource.FAILED] :
            print("  VM %s is alrady in final state"  %  vm_id)
            continue

        print("\nshutting down  %s "      %  cr.id)
        cr.destroy ()
        print("  state        : %s (%s)"  %  (state2str(cr.state), cr.state_detail))
    
    print() 
    sys.exit (0)


# --------------------------------------------------------------------------
else :

    usage ('invalid arguments')





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import re
import os
import sys
import time
import socket
import traceback

from ..url import Url
from ..    import exceptions as se


""" Provides an assortment of utilities """

_latencies = {}


# --------------------------------------------------------------------
#
def get_trace () :

    trace = sys.exc_info ()[2]

    if  trace :
        stack           = traceback.extract_tb  (trace)
        traceback_list  = traceback.format_list (stack)
        return "".join (traceback_list)

    else :
        stack           = traceback.extract_stack ()
        traceback_list  = traceback.format_list (stack)
        return "".join (traceback_list[:-1])


# --------------------------------------------------------------------
#
def host_is_local (host) :
    """ Returns True if the given host is the localhost
    """

    if  not host:
        return True

    elif host == 'localhost':
        return True

    else:
        sockhost = socket.gethostname()
        while sockhost:
            if host == sockhost:
                return True
            sockhost = '.'.join(sockhost.split('.')[1:])

    return False


# --------------------------------------------------------------------
#
def host_is_valid (host) :
    """ 
    Returns True if the given hostname can be resolved.
    We also test the reverse DNS lookup -- some seriously stupid and standard
    violating internet providers implement a DNS catchall -- the reverse lookup
    can catch that case in some cases (say that quickly 3 times!)
    """

    # FIXME: cache results so that further lookups are quick

    if  host_is_local (host) :
        return True

    try:
        _ = socket.gethostbyname (host)
        return True
    except :
        return False


# --------------------------------------------------------------------
#
def get_host_latency (host_url) :
    """ 
    This call measures the base tcp latency for a connection to the target
    host.  Note that port 22 is used for connection tests, unless the URL
    explicitly specifies a different port.  If the used port is blocked, the
    returned latency can be wrong by orders of magnitude.
    """

    try :
        # FIXME see comments to #62bebc9 -- this breaks for some cases, or is at
        # least annoying.  Thus we disable latency checking for the time being,
        # and return a constant assumed latency of 250ms (which approximately 
        # represents a random WAN link).
        return 0.25

        global _latencies

        if  host_url in _latencies :
            return _latencies[host_url]

        u = Url (host_url)

        if u.host : host = u.host
        else      : host = 'localhost'
        if u.port : port = u.port
        else      : port = 22  # FIXME: we should guess by protocol 

        # ensure host is valid
        if not host_is_valid(host):
            raise ValueError('invalid host')

        start = time.time ()

        s = socket.socket (socket.AF_INET, socket.SOCK_STREAM)
        s.connect  ((host, port))
        s.shutdown (socket.SHUT_RDWR)

        stop = time.time ()

        latency = stop - start

        _latencies[host_url] = latency

        return latency

    except :

        raise


# --------------------------------------------------------------------
#
def url_is_local (arg) :
    """ 
    Returns True if the given url points to localhost.

    We consider all URLs which explicitly define a port as non-local, because it
    looks like someone wants to explicitly use some protocol --
    `ssh://localost:2222/` is likely to point at an ssh tunnel.  

    If, however, the port matches the default port for the given protocol, we
    consider it local again -- `ssh://localhost:22/` is most likely a real local
    activity.

    Note that the schema set operates on substring level, so that we will accept
    port 22 for `ssh` and also for `sge+ssh` -- this may break in corner cases
    (altough I can't think of any right now).
    """

    u = Url (arg)

    if  not host_is_local (u.host) :
        return False

    # host is local, but what does the port indicate?
    if u.port and u.port > 0 :

        try :
            if  socket.getservbyport (u.port) in u.schema :
                # some non-default port is used -- consider remote
                return False
        except :
            # unknown service port --assume this is non-standard...
            return False

    # port is not set or points to default port for service
    return True


# --------------------------------------------------------------------
#
def url_is_relative (url_1) :
    """ an URL is considered relative if it only contains a path element, and
    that path element does not start with '/'.
    """

    u1 = Url (url_1)

    if  str (u1) == str(u1.path) :
        if  not u1.path :
            return True 
        elif u1.path[0] != '/' :
            return True

    return False


# --------------------------------------------------------------------
#
def url_get_dirname (url_1) :
    """ 
    Extract the directory part of the given URL's path element.  We consider
    everything up to the last '/' as directory.  That also holds for relative
    paths.
    """

    u1 = Url (url_1)
    p1 = u1.path

    return re.sub (r"[^/]*$", "", p1)


# --------------------------------------------------------------------
#
def url_get_filename (url_1) :
    """ 
    Extract the directory part of the given URL's path element.  We consider
    everything up to the last '/' as directory.  That also holds for relative
    paths.
    """

    u1 = Url (url_1)
    p1 = u1.path

    if '/' in p1 :
        return re.sub (r"^.*/(.*)$", "\1", p1)
    else :
        return p1


# --------------------------------------------------------------------
#
def url_normalize (url_1) :
    """ 
    The path element of the URL is normalized
    """

    ret      = Url (url_1)
    ret.path = os.path.normpath (ret.path)

    return ret


# --------------------------------------------------------------------
#
def url_make_absolute (url_1, url_2) :
    """ 
    URL1 is expected to only have a path
    Missing elements in url_1 are copied from url_2 -- but path stays the
    same.  Unless it is a relative path in the first place: then it is
    interpreted as relative to url_2.path, and is made absolute.
    protocol/port/user etc.
    """

    if  not isinstance(url_1, Url):
        url_1 = Url(url_1)

    if  not isinstance(url_2, Url):
        url_2 = Url(url_2)

    if  not url_is_compatible (url_1, url_2) :
        raise se.BadParameter("Cannot interpret url %s in the context of url %s"
                               % (url_2, url_1))

    # re-interpret path of url_2, using url_1 as base directory
    ret = Url (url_1)

    if  url_is_relative (url_2) :
        # note that we have no means if 'file://localhost/tmp/test.txt' refers
        # to a file or an directory -- so we have to trust that url_1 is
        # a dir...
        ret.path = url_1.path + '/' + url_2.path
    else :
        # absolute path, replace url path...
        ret.path = url_2.path


    return ret


# --------------------------------------------------------------------
#
def url_is_compatible (url_1, url_2) :
    """ 
    Returns True if the given urls point to the same host, using the same
    protocol/port/user etc.  If one of the URLs only contains a path, it is
    considered compatible with any other URL.
    """

    u1 = Url (url_1)
    u2 = Url (url_2)

    # if either one url only contains a path, it is compatible to anything.
    if os.path.normpath(u1.path) == os.path.normpath (str(u1)) : return True
    if os.path.normpath(u2.path) == os.path.normpath (str(u2)) : return True

    # more than path in both URLs -- check compatibility for all elements
    if u1.scheme   and     u2.scheme   and u1.scheme   != u2.scheme  : return False 
    if u1.host     and     u2.host     and u1.host     != u2.host    : return False
    if u1.port     and     u2.port     and u1.port     != u2.port    : return False
    if u1.username and     u2.username and u1.username != u2.username: return False
    if u1.password and     u2.password and u1.password != u2.password: return False

    if u1.scheme   and not u2.scheme  : return False 
    if u1.host     and not u2.host    : return False
    if u1.port     and not u2.port    : return False
    if u1.username and not u2.username: return False
    if u1.password and not u2.password: return False

    if u2.scheme   and not u1.scheme  : return False 
    if u2.host     and not u1.host    : return False
    if u2.port     and not u1.port    : return False
    if u2.username and not u1.username: return False
    if u2.password and not u1.password: return False

    # no differences detected (ignored fragments and query though)
    return True


# --------------------------------------------------------------------
#
def normalize_version (v) :
    """
    For a given version string (numeric only!), return an ordered tuple of
    integers, removing trailing zeros.  That tuple can then be used for
    comparison.
    """
    # parts = [int (x) for x in v.split (".")]
    # while parts[-1] == 0:
    #     parts.pop ()
    # return parts

    return tuple (v.split ("."))


# --------------------------------------------------------------------







__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Redis advert adaptor implementation """

from . import redis_namespace as rns

from ...url              import Url
from ...task             import Task
from ...                 import exceptions as rse
from ...                 import advert     as api
from ...utils            import misc       as rsumisc
from ...adaptors         import base       as a_base
from ...adaptors.cpi     import advert     as cpi
from ...adaptors.cpi     import decorators as cpi_decs


SYNC_CALL  = cpi_decs.SYNC_CALL
ASYNC_CALL = cpi_decs.ASYNC_CALL


###############################################################################
# adaptor info
#

_ADAPTOR_NAME          = 'radical.saga.adaptors.advert.redis'
_ADAPTOR_SCHEMAS       = ['redis']
_ADAPTOR_OPTIONS       = []
_ADAPTOR_CAPABILITIES  = {}

_ADAPTOR_DOC           = {
    'name'             : _ADAPTOR_NAME,
    'cfg_options'      : _ADAPTOR_OPTIONS,
    'capabilities'     : _ADAPTOR_CAPABILITIES,
    'description'      : 'The redis advert adaptor.',
    'details'          : """This adaptor interacts with a redis server to
                            implement the advert API semantics.""",
    'schemas'          : {'redis'  : 'redis nosql backend.'}
}

_ADAPTOR_INFO          = {
    'name'             : _ADAPTOR_NAME,
    'version'          : 'v0.2.beta',
    'schemas'          : _ADAPTOR_SCHEMAS,
    'cpis'             : [
                             {
                                 'type'  : 'radical.saga.advert.Directory',
                                 'class' : 'RedisDirectory'
                             },
                             {
                                 'type'  : 'radical.saga.advert.Entry',
                                 'class' : 'RedisEntry'
                             }
                         ]
}


###############################################################################
# The adaptor class

class Adaptor (a_base.Base):
    """
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self) :

        a_base.Base.__init__ (self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)

        # the adaptor *singleton* creates a (single) instance of a bulk handler
        # (BulkDirectory), which implements container_* bulk methods.
        self._bulk  = BulkDirectory ()
        self._redis = {}


    # --------------------------------------------------------------------------
    #
    def get_redis (self, url) :

        host     = None
        port     = 6379
        username = None
        password = None

        if url.host     : host     = url.host
        if url.port     : port     = url.port
        if url.username : username = url.username
        if url.password : password = url.password

        if username :
            if password :
                hash = "redis://%s:%s@%s:%d" % (username, password, host, port)
            else :
                hash = "redis://%s@%s:%d"    % (username,           host, port)
        else :
            if password :
                hash = "redis://%s@%s:%d"    % (password,       host, port)
            else :
                hash = "redis://%s:%d"       %                 (host, port)

        if hash not in self._redis :
            self._redis[hash] = rns.redis_ns_server (url)

        return self._redis[hash]


    # --------------------------------------------------------------------------
    #
    def sanity_check (self) :
        # nothing to check for, redis entry system should always be accessible
        pass


###############################################################################
#
class BulkDirectory (cpi.Directory) :
    """
    Well, this implementation can handle bulks, but cannot optimize them.
    We leave that code here anyway, for demonstration -- but those methods
    are also provided as fallback, and are thusly used if the adaptor does
    not implement the bulk container_* methods at all.
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self) :
        pass


    # --------------------------------------------------------------------------
    #
    def container_wait (self, tasks, mode, timeout) :

        if timeout >= 0 :
            raise rse.BadParameter ("Cannot handle timeouts > 0")
        for task in tasks :
            task.wait ()


    # --------------------------------------------------------------------------
    #
    def container_cancel (self, tasks, timeout) :

        for task in tasks :
            task.cancel (timeout)


    # --------------------------------------------------------------------------
    #
    def container_copy (self, tasks) :
        """
        A *good* implementation would dig the entry copy operations from the
        tasks, and run them in a bulk -- we can't do that, so simply *run* the
        individual tasks, falling back to the default non-bulk asynchronous copy
        operation...
        """
        for task in tasks :
            task.run ()


    # the container methods for the other calls are obviously similar, and left
    # out here.  The :class:`saga.task.Container` implementation will fall back
    # to the non-bulk async calls for all then.


###############################################################################
#
class RedisDirectory (cpi.Directory) :

    # --------------------------------------------------------------------------
    #
    def __init__ (self, api, adaptor) :

        self._cpi_base = super  (RedisDirectory, self)
        self._cpi_base.__init__ (api, adaptor)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance (self, adaptor_state, url, flags, session) :

        self._url       = rsumisc.url_normalize (url)
        self._flags     = flags
        self._container = self._adaptor._bulk

        self._set_session (session)
        self._init_check  ()

        return self.get_api ()


    # --------------------------------------------------------------------------
    #
    @ASYNC_CALL
    def init_instance_async (self, adaptor_state, url, flags, session, ttype) :

        self._url     = rsumisc.url_normalize (url)
        self._flags   = flags

        self._set_session (session)

        c = {'url'     : self._url,
             'flags'   : self._flags }

        return Task (self, 'init_instance', c, ttype)


    # --------------------------------------------------------------------------
    #
    def _init_check (self) :

        self._r       = self._adaptor.get_redis (self._url)
        self._nsdir   = rns.redis_ns_entry.opendir(self._r, self._url.path,
                                                  self._flags)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def attribute_getter (self, key) :

        try :
            return self._nsdir.get_key (key)

        except Exception as e :
            self._logger.error ("get_key failed: %s" % e)
            raise


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def attribute_setter (self, key, val) :

        try :
            self._nsdir.set_key (key, val)

        except Exception as e :
            self._logger.error ("set_key failed: %s" % e)
            raise


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def attribute_lister (self) :

        data = self._nsdir.get_data ()

        for key in list(data.keys ()) :
            self._api ()._attributes_i_set (key, data[key], self._api ()._UP)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def attribute_caller (self, key, id, cb) :

        self._nsdir.manage_callback (key, id, cb, self.get_api ())


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url (self) :

        return self._url


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir (self, name) :

        try :
            api.Directory (rsumisc.url_make_absolute (self._url, name))
        except Exception:
            return False

        return True


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list (self, pattern, flags) :

        if  pattern :
            raise rse.BadParameter ("pattern for list() not supported")


        ret = []

        if  not flags :

            ret = self._nsdir.list ()


        elif flags == api.RECURSIVE :

            # ------------------------------------------------------------------
            def get_kids (path) :

                d    = api.Directory (path)
                kids = d.list ()

                for kid in kids :

                    kid_url      = self._url
                    kid_url.path = kid

                    if  d.is_dir (kid_url) :
                        get_kids (kid_url)

                    ret.append (kid)
            # ------------------------------------------------------------------

            get_kids (self._url)


        else :
            raise rse.BadParameter ("list() only supports the RECURSIVE flag")


        return ret


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def change_dir (self, tgt) :

        # backup state
        orig_url = self._url

        try :
            if  not rsumisc.url_is_compatible (tgt, self._url) :
                raise rse.BadParameter("cannot chdir to %s, leaves namespace"
                                      % tgt)

            self._url = rsumisc.url_make_absolute (tgt, self._url)
            self._init_check ()


        finally :
            # restore state on error
            self._url = orig_url



    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def open (self, url, flags) :

        if not url.scheme and not url.host :
            url = Url (str(self._url) + '/' + str(url))

        return api.Entry (url, flags, self._session, _adaptor=self._adaptor)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def open_dir (self, url, flags) :

        if not url.scheme and not url.host :
            url = Url (str(self._url) + '/' + str(url))

        return api.Directory (url, flags, self._session, _adaptor=self._adaptor)


  # ##################################################################
  # # FIXME: all below
  # @SYNC_CALL
  # def copy (self, source, target, flags) :
  #     return
  #
  #     src_url = Url (source)
  #     src     = src_url.path
  #     tgt_url = Url (target)
  #     tgt     = tgt_url.path
  #
  #
  #     if src_url.schema :
  #         if not src_url.schema.lower () in _ADAPTOR_SCHEMAS :
  #             raise rse.BadParameter ("Cannot handle url %s (not redis)"
  #                                    % source)
  #
  #     if tgt_url.schema :
  #         if not tgt_url.schema.lower () in _ADAPTOR_SCHEMAS :
  #             raise rse.BadParameter ("Cannot handle url %s (not redis)"
  #                                    % target)
  #
  #
  #     # make paths absolute
  #     if src[0] != '/'  :  src = "%s/%s"   % (os.path.dirname (src), src)
  #     if tgt[0] != '/'  :  tgt = "%s/%s"   % (os.path.dirname (src), tgt)
  #
  #     shutil.copy2 (src, tgt)
  #
  #
  # @ASYNC_CALL
  # def copy_async (self, src, tgt, flags, ttype) :
  #
  #     c = { 'src'     : src,
  #           'tgt'     : tgt,
  #           'flags'   : flags }
  #
  #     return apip_task.Task (self, 'copy', c, ttype)
  #
  #
  #
  # def task_wait (self, task, timout) :
  #     # FIXME: our task_run moves all tasks into DONE state... :-/
  #     pass


######################################################################
#
# entry adaptor class
#
class RedisEntry (cpi.Entry) :

    # --------------------------------------------------------------------------
    #
    def __init__ (self, api, adaptor) :

        self._cpi_base = super  (RedisEntry, self)
        self._cpi_base.__init__ (api, adaptor)


    # --------------------------------------------------------------------------
    #
    def _dump (self) :

        self._logger.debug ("url  : %s"  % self._url)
        self._logger.debug ("flags: %s"  % self._flags)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance (self, adaptor_state, url, flags, session) :

        self._url     = url
        self._flags   = flags

        self._set_session (session)
        self._init_check  ()

        return self


    # --------------------------------------------------------------------------
    #
    def _init_check (self) :

        self._r       = self._adaptor.get_redis (self._url)
        self._nsentry = rns.redis_ns_entry.open (self._r, self._url.path,
                                                 self._flags)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def attribute_getter (self, key) :

        return self._nsentry.get_key (key)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def attribute_setter (self, key, val) :

        return self._nsentry.set_key (key, val)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def attribute_lister (self) :

        data = self._nsentry.get_data ()

        for key in list(data.keys ()) :
            self._api ()._attributes_i_set (key, data[key], self._api ()._UP)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def attribute_caller (self, key, id, cb) :

        return self._nsentry.manage_callback (key, id, cb, self.get_api ())


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url (self) :

        return self._url


  # ##################################################################
  # # FIXME: all below
  # @SYNC_CALL
  # def copy_self (self, target, flags) :
  #
  #     tgt_url = Url (target)
  #     tgt     = tgt_url.path
  #     src     = self._url.path
  #
  #     if tgt_url.schema :
  #         if not tgt_url.schema.lower () in _ADAPTOR_SCHEMAS :
  #             raise rse.BadParameter ("Cannot handle url %s (not redis)"
  #                                    %  target)
  #
  #     if not rsumisc.url_is_redis (tgt_url) :
  #         raise rse.BadParameter ("Cannot handle url %s (not redis)"
  #                                %  target)
  #
  #     # make path absolute
  #     if tgt[0] != '/'  :  tgt = "%s/%s"   % (os.path.dirname (src), tgt)
  #
  #     shutil.copy2 (src, tgt)


# ------------------------------------------------------------------------------






__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .decorators import CPI_SYNC_CALL  as SYNC
from .decorators import CPI_ASYNC_CALL as ASYNC


class Async (object) :
    
    @SYNC
    def task_run    (self, task)          : pass

    @SYNC
    def task_wait   (self, task, timeout) : pass

    @SYNC
    def task_cancel (self, task)          : pass








__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import os
import pwd
import time
import getpass
import threading as mt

import radical.utils           as ru

from   . import misc           as sumisc
from   . import pty_process    as supp
from   . import pty_exceptions as ptye
from   ..import exceptions     as rse

from   ..session import Session
from   ..url     import Url



# ------------------------------------------------------------------------------
#
# ssh options:
#   -e none         : no escape character
#   -M              : master mode for connection sharing
#   -S control_path : slave mode for connection sharing
#   -t              : force pty allocation
#   -x              : disable x11 forwarding
#
#   ServerAliveInterval
#   CheckHostIP no
#   ConnectTimeout
#   ControlMaster  yes | no | no ...
#   ControlPath    $BASE/ssh_control_%n_%p.$$.sock
#                  %r (remote id)? would need inspection
#   ControlPersist 100  : close master after 100 seconds idle
#   EscapeChar     none : transparent for binary data
#   TCPKeepAlive   yes  : detect connection failure
#
#   LoginGraceTime seconds : disconnect if no login after n seconds
#
# ------------------------------------------------------------------------------

# these arrays help to map requested client schemas to master schemas
_SCHEMAS_SH  = ['sh', 'fork', 'local', 'file']
_SCHEMAS_SSH = ['ssh', 'scp', 'sftp']
_SCHEMAS_GSI = ['gsissh', 'gsiscp', 'gsisftp']

_SCHEMAS = _SCHEMAS_SH + _SCHEMAS_SSH + _SCHEMAS_GSI

# FIXME: '-o ControlPersist' is only supported for newer ssh versions.  We
# should add detection, and enable that if available -- for now, just diable it.
#
# FIXME: we should use '%n' instead of '%h', but that is not supported by older
# ssh versions...

# ssh master/slave flag magic # FIXME: make timeouts configurable
_SSH_FLAGS_MASTER   = "-o ControlMaster=%(share_mode)s -o ControlPath=%(ctrl)s -o TCPKeepAlive=no  -o ServerAliveInterval=10 -o ServerAliveCountMax=20 %(connect_timeout)s"
_SSH_FLAGS_SLAVE    = "-o ControlMaster=%(share_mode)s -o ControlPath=%(ctrl)s -o TCPKeepAlive=no  -o ServerAliveInterval=10 -o ServerAliveCountMax=20 %(connect_timeout)s"
_SCP_FLAGS          = ""
_SFTP_FLAGS         = ""

# FIXME: right now, we create a shell connection as master --
# but a master does not actually need a shell, as it is never really
# used to run commands...
_SCRIPTS = {
    'ssh' : {
        'master'       : '%(ssh_env)s "%(ssh_exe)s" %(ssh_args)s %(m_flags)s %(host_str)s',
        'shell'        : '%(ssh_env)s "%(ssh_exe)s" %(ssh_args)s %(s_flags)s %(host_str)s',
        'copy_is_posix': True
    },
    'scp' : {
        'copy_to'      : '%(scp_env)s "%(scp_exe)s" %(scp_args)s %(s_flags)s %(cp_flags)s "%(src)s" "%(scp_root)s%(tgt)s"',
        'copy_from'    : '%(scp_env)s "%(scp_exe)s" %(scp_args)s %(s_flags)s %(cp_flags)s "%(scp_root)s%(src)s" "%(tgt)s"',
        'copy_to_in'   : '',
        'copy_from_in' : '',
        'copy_is_posix': False
    },
    'sftp' : {
        'copy_to'      : '%(sftp_env)s "%(sftp_exe)s" %(sftp_args)s %(s_flags)s %(host_str)s',
        'copy_from'    : '%(sftp_env)s "%(sftp_exe)s" %(sftp_args)s %(s_flags)s %(host_str)s',
        'copy_to_in'   : 'mput %(cp_flags)s "%(src)s" "%(tgt)s"',
        'copy_from_in' : 'mget %(cp_flags)s "%(src)s" "%(tgt)s"',
        'copy_is_posix': False
    },
    'sh' : {
        'master'       : '%(sh_env)s "%(sh_exe)s"  %(sh_args)s',
        'shell'        : '%(sh_env)s "%(sh_exe)s"  %(sh_args)s',
        'copy_to'      : '%(sh_env)s "%(sh_exe)s"  %(sh_args)s',
        'copy_from'    : '%(sh_env)s "%(sh_exe)s"  %(sh_args)s',
        'copy_to_in'   : 'cd ~ && "%(cp_exe)s" -v %(cp_flags)s "%(src)s" "%(tgt)s"',
        'copy_from_in' : 'cd ~ && "%(cp_exe)s" -v %(cp_flags)s "%(src)s" "%(tgt)s"',
        'copy_is_posix': True
    }
}


# ------------------------------------------------------------------------------
#
class PTYShellFactory (object, metaclass=ru.Singleton) :
    """
    This is the place where all master and all client shell connections get
    created.  But also, this factory maintains a registry of master connections,
    to quickly spawn slave connections for any customer w/o repeated
    authorization overhead.  Masters are unique per
    a host/user/port/context/shell_type hash.

    Any ssh master connection in this registry can idle, and may thus shut down
    after ``ControlPersist`` seconds (see options).

    data model::


      self.registry
        |
        +-- "host[:port]"
        |   |
        |   +-- str(context)
        |   |   |
        |   |   +-- "shell_type" (ssh)
        |   |   |   |
        |   |   |   +-- pty_process (gc_timeout)
        |   |   |   +-- shell_initialize()
        |   |   |   +-- shell_finalize()
        |   |   |   +-- shell_alive()
        |   |   |
        |   |   +-- ...
        |   |
        |   +-- ...
        |
        +-- ...

    When Slave connections are requested, a suitable master connection is looked
    for and used.  'Suitable' means: ssh master for scp and sftp slaves; gsissh
    for gsiscp and gsisftp slaves; and sh master for file slaves

    """


    # --------------------------------------------------------------------------
    #
    def __init__ (self, cfg=None):

        self.logger     = ru.Logger('radical.saga.pty')
        self.rlock      = mt.RLock()
        self.registry   = dict()

        name = None
        if isinstance(cfg, str):
            name = cfg
            cfg  = None

        self.cfg = ru.Config('radical.saga.session', name=name, cfg=cfg)
        self.cfg = self.cfg.pty


    # --------------------------------------------------------------------------
    #
    def initialize (self, url, session=None, prompt=None, logger=None, cfg=None,
            posix=True, interactive=True) :

        with self.rlock :

            # make sure we have a valid url type
            url = Url (url)

            if  not prompt :
                prompt = "^(.*[\$#%>\]])\s*$"

            if  not logger :
                logger = self.logger


            # collect all information we have/need about the requested master
            # connection
            info = self._create_master_entry (url, session, prompt, logger, cfg,
                    posix, interactive)

            # we got master info - register the master, and create the instance!
            type_s = str(info['shell_type'])
            user_s = str(info['user'])
            host_s = str(info['host_str'])

            # Now, if we don't have that master, yet, we need to instantiate it
            if host_s not in self.registry                 : self.registry[host_s] = {}
            if user_s not in self.registry[host_s]         : self.registry[host_s][user_s] = {}
            if type_s not in self.registry[host_s][user_s] :

                # new master: create an instance, and register it
                m_cmd = info['scripts'][info['shell_type']]['master'] % info

                logger.debug ("open master pty for [%s] [%s] %s: %s'"
                                % (type_s, host_s, user_s, m_cmd))

                info['pty'] = supp.PTYProcess (m_cmd, cfg, logger)
                if not info['pty'].alive () :
                    raise rse.NoSuccess._log (logger,
                          "Shell not connected to %s" % info['host_str'])

                # authorization, prompt setup, etc.  Initialize as shell if not
                # explicitly marked as non-posix shell
                self._initialize_pty(info['pty'], info)

                # master was created - register it
                self.registry[host_s][user_s][type_s] = info


            else :
                # we already have a master: make sure it is alive, and restart
                # as needed
                info = self.registry[host_s][user_s][type_s]

                if  not info['pty'].alive (recover=True) :
                    raise rse.IncorrectState._log (logger,
                          "Lost shell connection to %s" % info['host_str'])

            return info


    # --------------------------------------------------------------------------
    #
    def _initialize_pty (self, pty_shell, info, posix=None) :

        # posix: only for posix shells we use prompt triggers.  sftp for example
        # does not deal well with triggers (no printf).

        with self.rlock :

            # import pprint
            # pprint.pprint(info)

            shell_pass = info['pass']
            key_pass   = info['key_pass']
            prompt     = info['prompt']
            logger     = info['logger']
            latency    = info['latency']
            timeout    = info['ssh_timeout']

            pty_shell.latency = latency

            if posix is None:
                posix = info['posix']

            # if we did not see a decent prompt within 'delay' time, something
            # went wrong.  Try to prompt a prompt (duh!)  Delay should be
            # minimum 0.1 second (to avoid flooding of local shells), and at
            # maximum 1 second (to keep startup time reasonable)
            # most one second.  We try to get within that range with 10*latency.
            delay = min (1.0, max (0.1, 10 * latency))

            try :
                prompt_patterns = [
                     "[Pp]assword:\s*$",              # password   prompt
                     "Enter passphrase for .*:\s*$",  # passphrase prompt
                     "Token_Response.*:\s*$",         # passtoken  prompt
                     "Enter PASSCODE:$",              # RSA SecureID
                     "want to continue connecting",   # hostkey confirmation
                     ".*HELLO_\\d+_SAGA$",            # prompt detection helper
                     prompt]                          # greedy native prompt

                # use a very aggressive, but portable prompt setting scheme.
                # Error messages may appear for tcsh and others.  Excuse
                # non-posix shells
                if posix:
                    pty_shell.write(" export PROMPT_COMMAND='' PS1='$';"
                                    " set prompt='$'\n")

                # find a prompt
                n, match = pty_shell.find (prompt_patterns, delay)

                # this loop will run until we finally find the shell prompt, or
                # if we think we have tried enough and give up.  On success
                # we'll try to set a different prompt, and when we found that,
                # too, we exit the loop and are be ready to running shell

                # commands.
                retries       = 0
                retry_trigger = True
                used_trigger  = False
                found_trigger = ""
                time_start    = time.time()

                while True :

                    # ----------------------------------------------------------
                    if n is None :

                        # we found none of the prompts, yet, and need to try
                        # again.  But to avoid hanging on invalid prompts, we
                        # print 'HELLO_x_SAGA', and search for that one, too.
                        # We actually do 'printf HELLO_%d_SAGA x' so that the
                        # pattern only appears in the result, not in the
                        # command...

                        if time.time() - time_start > timeout:
                            raise rse.NoSuccess("Could not detect shell prompt (timeout)")

                        # make sure we retry a finite time...
                        retries += 1

                        if  not retry_trigger :
                            # just waiting for the *right* trigger or prompt,
                            # don't need new ones...
                            continue

                        if posix:
                            # use a very aggressive, but portable prompt
                            # setting scheme
                            pty_shell.write (" export PROMPT_COMMAND='' PS1='$' > /dev/null 2>&1 || set prompt='$'\n")
                            pty_shell.write (" printf 'HELLO_%%d_SAGA\\n' %d\n" % retries)
                            used_trigger = True

                        # FIXME:  consider better timeout
                        n, match = pty_shell.find (prompt_patterns, delay)


                    # --------------------------------------------------------------
                    elif n == 0 :
                        logger.info ("got password prompt")
                        if  not shell_pass :
                            raise rse.AuthenticationFailed ("prompted for unknown password (%s)" \
                                                          % match)

                        pty_shell.write ("%s\n" % shell_pass, nolog=True)
                        n, match = pty_shell.find (prompt_patterns, delay)


                    # --------------------------------------------------------------
                    elif n == 1 :
                        logger.info ("got passphrase prompt : %s" % match)

                        start = match.find("'", 0)
                        end   = match.find("'", start + 1)

                        if start == -1 or end == -1 :
                            raise rse.AuthenticationFailed(
                                    "could not extract key name (%s)" % match)

                        key = match[start + 1:end]

                        if  key not in key_pass    :
                            raise rse.AuthenticationFailed(
                                "prompted for unknown key password (%s)" % key)

                        pty_shell.write ("%s\n" % key_pass[key], nolog=True)
                        n, match = pty_shell.find (prompt_patterns, delay)


                    # --------------------------------------------------------------
                    elif n == 2 or n == 3:
                        logger.info ("got token prompt")
                        import getpass
                        token = getpass.getpass ("enter token: ")
                        pty_shell.write ("%s\n" % token.strip(), nolog=True)
                        n, match = pty_shell.find (prompt_patterns, delay)


                    # --------------------------------------------------------------
                    elif n == 4:
                        logger.info ("got hostkey prompt")
                        pty_shell.write ("yes\n")
                        n, match = pty_shell.find (prompt_patterns, delay)


                    # --------------------------------------------------------------
                    elif n == 5:

                        # one of the trigger commands got through -- we can now
                        # hope to find the prompt (or the next trigger...)
                        logger.debug("got shell prompt trigger (%s) (%s)"
                                    % (n, match))

                        found_trigger = match
                        retry_trigger = False
                        n, match = pty_shell.find (prompt_patterns, delay)
                        continue


                    # --------------------------------------------------------------
                    elif n == 6 :

                        logger.debug ("got initial shell prompt (%s) (%s)"
                                % (n, match))

                        if  retries :
                            if  used_trigger :
                                # we already sent triggers -- so this match is
                                # only useful if saw the *correct* shell prompt
                                # trigger first
                                trigger = "HELLO_%d_SAGA" % retries

                                if  trigger not in found_trigger :

                                    logger.debug ("waiting for prompt trigger %s: (%s) (%s)"
                                               % (trigger, n, match))
                                    # but more retries won't help...
                                    retry_trigger = False
                                    attempts      = 0
                                    n             = None

                                    while not n :

                                        attempts += 1
                                        n, match = pty_shell.find(prompt_patterns, delay)

                                        if  not n :
                                            if  attempts == 1 :
                                                if posix:
                                                    pty_shell.write (" printf 'HELLO_%%d_SAGA\\n' %d\n" % retries)

                                            if  attempts > 100 :
                                                raise rse.NoSuccess ("Could not detect shell prompt (timeout)")

                                    continue


                        logger.debug ("Got initial shell prompt (%s) (%s)" % (n, match))
                        # we are done waiting for a prompt
                        break

            except Exception as e :
                logger.exception(e)
                raise ptye.translate_exception (e) from e


    # --------------------------------------------------------------------------
    #
    def get_cp_slave (self, s_cmd, info, posix=None) :

        with self.rlock :

            if posix is None:
                posix = info.get('copy_is_posix')

          # print('> -- new cp  shell to %s' % s_cmd)

            cp_slave = supp.PTYProcess (s_cmd, self.cfg, info['logger'])
            self._initialize_pty (cp_slave, info, posix)

            return cp_slave


    # --------------------------------------------------------------------------
    #
    def run_shell (self, info) :
        """
        This initiates a master connection.  If there is a suitable master
        connection in the registry, it is re-used, and no new master connection
        is created.  If needed, the existing master connection is revived.
        """

        with self.rlock :

            s_cmd = info['scripts'][info['shell_type']]['shell'] % info

            # at this point, we do have a valid, living master
            sh_slave = supp.PTYProcess (s_cmd, self.cfg, info['logger'])

            # authorization, prompt setup, etc
            self._initialize_pty (sh_slave, info)

            return sh_slave


    # --------------------------------------------------------------------------
    #
    def _which(self, cmd):

        ret = ru.which(cmd)
        if not ret:
            raise RuntimeError('cmd %s not found' % cmd)
        return ret


    # --------------------------------------------------------------------------
    #
    def _create_master_entry (self, url, session, prompt, logger, cfg, posix,
            interactive) :
        # FIXME: cache 'which' results, etc
        # FIXME: check 'which' results

        with self.rlock :

            info = {'posix' : posix}

            # get and evaluate session config
            if  not session :
                session = Session (default=True)

            info['ssh_copy_mode']  = cfg['ssh_copy_mode']
            info['ssh_share_mode'] = cfg['ssh_share_mode']
            info['ssh_timeout']    = cfg['ssh_timeout']

            logger.info ("ssh copy  mode set to '%s'" % info['ssh_copy_mode' ])
            logger.info ("ssh share mode set to '%s'" % info['ssh_share_mode'])
            logger.info ("ssh timeout    set to '%s'" % info['ssh_timeout'])


            # fill the info dict with details for this master channel, and all
            # related future slave channels
            info['schema']    = url.schema.lower ()
            info['host_str']  = url.host
            info['prompt']    = prompt
            info['logger']    = logger
            info['url']       = url
            info['pass']      = ""
            info['key_pass']  = {}
            info['scripts']   = _SCRIPTS

            if  not info['schema'] :
                info['schema'] = 'local'


            # find out what type of shell we have to deal with
            if  info['schema'] in _SCHEMAS_SSH :
                info['shell_type'] = "ssh"
                info['copy_mode']  = info['ssh_copy_mode']
                info['share_mode'] = info['ssh_share_mode']
                info['ssh_exe']    = self._which ("ssh")
                info['scp_exe']    = self._which ("scp")
                info['sftp_exe']   = self._which ("sftp")

            elif info['schema'] in _SCHEMAS_GSI :
                info['shell_type'] = "ssh"
                info['copy_mode']  = info['ssh_copy_mode']
                info['share_mode'] = info['ssh_share_mode']
                info['ssh_exe']    = self._which ("gsissh")
                info['scp_exe']    = self._which ("gsiscp")
                info['sftp_exe']   = self._which ("gsisftp")

            elif info['schema'] in _SCHEMAS_SH :
                info['shell_type'] = "sh"
                info['copy_mode']  = "sh"
                info['share_mode'] = "auto"
                info['sh_env']     = "/usr/bin/env TERM=vt100 PROMPT_COMMAND='' PS1='PROMPT-$?->'"
                info['cp_env']     = "/usr/bin/env TERM=vt100 PROMPT_COMMAND='' PS1='PROMPT-$?->'"
                info['scp_root']   = "/"

                if interactive: info['sh_args'] = "-i"
                else          : info['sh_args'] = ""

                if  "SHELL" in os.environ :
                    info['sh_exe'] =  self._which (os.environ["SHELL"])
                    info['cp_exe'] =  self._which ("cp")
                else :
                    info['sh_exe'] =  self._which ("sh")
                    info['cp_exe'] =  self._which ("cp")

            else :
                raise rse.BadParameter._log (self.logger,
                          "cannot handle schema '%s://'" % url.schema)


            # If an SSH timeout has been specified set up the ConnectTimeout
            # string
            if info['ssh_timeout']:
                info['ssh_connect_timeout'] = ('-o ConnectTimeout=%s'
                    % int(float(info['ssh_timeout'])))
            else:
                info['ssh_connect_timeout'] = ''

            # depending on type, create command line (args, env etc)
            #
            # We always set term=vt100 to avoid ansi-escape sequences in the
            # prompt and elsewhere.  Also, we have to make sure that the shell
            # is an interactive login shell, so that it interprets the users
            # startup files, and reacts on commands.

            try :
                info['latency'] = sumisc.get_host_latency (url)

                # FIXME: note that get_host_latency is considered broken (see
                # saga/utils/misc.py line 73), and will return a constant 250ms.

            except Exception as e :
                info['latency'] = 1.0  # generic value assuming slow link
                info['logger'].warning("Could not contact host '%s': %s"
                                      % (url, e))

            if  info['shell_type'] == "sh" :

                info['sh_env'] = "/usr/bin/env TERM=vt100 "  # avoid ansi escapes

              # if not sumisc.host_is_local (url.host) :
              #     raise rse.BadParameter._log (self.logger,
              #             "expect local host for '%s://', not '%s'"
              #             % (url.schema, url.host))

                if  'user' in info and info['user'] :
                    pass
                else :
                    info['user'] = getpass.getuser ()

            else :
                # avoid ansi escapes
                info['ssh_env']   = "/usr/bin/env TERM=vt100 "
                info['scp_env']   = "/usr/bin/env TERM=vt100 "
                info['sftp_env']  = "/usr/bin/env TERM=vt100 "
                info['ssh_args']  = "-t "  # force pty
                info['scp_args']  = _SCP_FLAGS
                info['sftp_args'] = _SFTP_FLAGS

                if  session :

                    for context in session.contexts :

                        # ssh can also handle UserPass contexts, and ssh type
                        # contexts.  gsissh can handle the same, but also X509
                        # contexts.

                        if  context.type.lower () == "ssh" :
                            if  info['schema'] in _SCHEMAS_SSH + _SCHEMAS_GSI :

                                if  context.attribute_exists ("user_id") and \
                                    context.user_id :
                                    info['user']  = context.user_id

                                if  context.attribute_exists ("user_key") and \
                                    context.user_key  :
                                    info['ssh_args']  += "-o IdentityFile=%s " % context.user_key
                                    info['scp_args']  += "-o IdentityFile=%s " % context.user_key
                                    info['sftp_args'] += "-o IdentityFile=%s " % context.user_key

                                    if  context.attribute_exists("user_pass") \
                                        and context.user_pass:
                                        info['key_pass'][context.user_key] = context.user_pass

                        if  context.type.lower () == "userpass" :
                            if  info['schema'] in _SCHEMAS_SSH + _SCHEMAS_GSI :
                                if  context.attribute_exists ("user_id") and \
                                    context.user_id :
                                    info['user'] = context.user_id
                                if  context.attribute_exists ("user_pass") and \
                                    context.user_pass :
                                    info['pass'] = context.user_pass

                        if  context.type.lower () == "x509" :
                            if  info['schema'] in _SCHEMAS_GSI :

                                if  context.attribute_exists ("user_proxy")  and   context.user_proxy :
                                    info['ssh_env']   += "X509_USER_PROXY='%s' " % context.user_proxy
                                    info['scp_env']   += "X509_USER_PROXY='%s' " % context.user_proxy
                                    info['sftp_env']  += "X509_USER_PROXY='%s' " % context.user_proxy

                                if  context.attribute_exists ("user_cert")   and  context.user_cert :
                                    info['ssh_env']   += "X509_USER_CERT='%s' " % context.user_cert
                                    info['scp_env']   += "X509_USER_CERT='%s' " % context.user_cert
                                    info['sftp_env']  += "X509_USER_CERT='%s' " % context.user_cert

                                if  context.attribute_exists ("user_key")    and  context.user_key :
                                    info['ssh_env']   += "X509_USER_key='%s' "  % context.user_key
                                    info['scp_env']   += "X509_USER_key='%s' "  % context.user_key
                                    info['sftp_env']  += "X509_USER_key='%s' "  % context.user_key

                                if  context.attribute_exists ("cert_repository") and context.cert_repository :
                                    info['ssh_env']   += "X509_CERT_DIR='%s' "  % context.cert_repository
                                    info['scp_env']   += "X509_CERT_DIR='%s' "  % context.cert_repository
                                    info['sftp_env']  += "X509_CERT_DIR='%s' "  % context.cert_repository

                if url.port and url.port != -1 :
                    info['ssh_args']  += "-o Port=%d " % int(url.port)
                    info['scp_args']  += "-o Port=%d " % int(url.port)
                    info['sftp_args'] += "-o Port=%d " % int(url.port)


                # all ssh based shells allow for user_id and user_pass from
                # contexts -- but the data given in the URL take precedence

                if url.username   :  info['user'] = url.username
                if url.password   :  info['pass'] = url.password

                ctrl_user = pwd.getpwuid (os.getuid ()).pw_name
                ctrl_base = "/tmp/saga_ssh_%s" % ctrl_user


                if  'user' in info and info['user'] :
                    info['host_str'] = "%s@%s"              % (info['user'], info['host_str'])
                    info['ctrl']     = "%s_%%h_%%p.%s.ctrl" % (ctrl_base, info['user'])
                else :
                    info['user'] = getpass.getuser ()
                    info['ctrl'] = "%s_%%h_%%p.ctrl" % (ctrl_base)

                info['m_flags']  = _SSH_FLAGS_MASTER % ({'share_mode' : info['share_mode'],
                                                         'ctrl'       : info['ctrl'],
                                                         'connect_timeout': info['ssh_connect_timeout']})
                info['s_flags']  = _SSH_FLAGS_SLAVE  % ({'share_mode' : info['share_mode'],
                                                         'ctrl'       : info['ctrl'],
                                                         'connect_timeout': info['ssh_connect_timeout']})

                logger.debug('SSH Connection M_FLAGS: %s' % info['m_flags'])
                logger.debug('SSH Connection S_FLAGS: %s' % info['s_flags'])

                # we want the userauth and hostname parts of the URL, to get the
                # scp-scope fs root.
                info['scp_root']  = ""
                has_auth          = False
                if  url.username :
                    info['scp_root'] += url.username
                    has_auth          = True
                if  url.password :
                    info['scp_root'] += ":"
                    info['scp_root'] += url.password
                    has_auth          = True
                if  has_auth :
                    info['scp_root'] += "@"
                info['scp_root']     += "%s:" % url.host

                # FIXME: port needs to be handled as parameter
              # if  url.port :
              #     info['scp_root'] += ":%d" % url.port


            # keep all collected info in the master dict, and return it for
            # registration
            return info


# ------------------------------------------------------------------------------
